# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/stream_data_source__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import data_source_config__client_pb2 as tecton__proto_dot_args_dot_data__source__config____client__pb2
from tecton_proto.args import transformation__client_pb2 as tecton__proto_dot_args_dot_transformation____client__pb2
from tecton_proto.args import user_defined_function__client_pb2 as tecton__proto_dot_args_dot_user__defined__function____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2
from tecton_proto.common import spark_schema__client_pb2 as tecton__proto_dot_common_dot_spark__schema____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2tecton_proto/data/stream_data_source__client.proto\x12\x11tecton_proto.data\x1a\x32tecton_proto/args/data_source_config__client.proto\x1a.tecton_proto/args/transformation__client.proto\x1a\x35tecton_proto/args/user_defined_function__client.proto\x1a(tecton_proto/common/schema__client.proto\x1a.tecton_proto/common/spark_schema__client.proto\"\\\n\x0fKinesisDSConfig\x12\x1f\n\x0bstream_name\x18\x01 \x01(\tR\nstreamName\x12\x16\n\x06region\x18\x02 \x01(\tR\x06regionJ\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06\"\xfd\x02\n\rKafkaDSConfig\x12+\n\x11\x62ootstrap_servers\x18\x01 \x01(\tR\x10\x62ootstrapServers\x12\x16\n\x06topics\x18\x02 \x01(\tR\x06topics\x12\x32\n\x15ssl_keystore_location\x18\x07 \x01(\tR\x13sslKeystoreLocation\x12\x44\n\x1fssl_keystore_password_secret_id\x18\x08 \x01(\tR\x1bsslKeystorePasswordSecretId\x12\x36\n\x17ssl_truststore_location\x18\t \x01(\tR\x15sslTruststoreLocation\x12H\n!ssl_truststore_password_secret_id\x18\n \x01(\tR\x1dsslTruststorePasswordSecretId\x12+\n\x11security_protocol\x18\x0b \x01(\tR\x10securityProtocol\"c\n\x1dSparkStreamDataSourceFunction\x12\x42\n\x08\x66unction\x18\x01 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x08\x66unction\"\xd6\x06\n\x10StreamDataSource\x12T\n\x13kinesis_data_source\x18\x01 \x01(\x0b\x32\".tecton_proto.data.KinesisDSConfigH\x00R\x11kinesisDataSource\x12N\n\x11kafka_data_source\x18\x0b \x01(\x0b\x32 .tecton_proto.data.KafkaDSConfigH\x00R\x0fkafkaDataSource\x12o\n\x1aspark_data_source_function\x18\x0c \x01(\x0b\x32\x30.tecton_proto.data.SparkStreamDataSourceFunctionH\x00R\x17sparkDataSourceFunction\x12J\n\x0bpush_source\x18\r \x01(\x0b\x32\'.tecton_proto.data.PushDataSourceConfigH\x00R\npushSource\x12\x43\n\x0cspark_schema\x18\t \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0bsparkSchema\x12Z\n\x15raw_stream_translator\x18\x03 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x13rawStreamTranslator\x12<\n\x1a\x64\x65\x64uplication_column_names\x18\n \x03(\tR\x18\x64\x65\x64uplicationColumnNames\x12\x1f\n\x0btime_column\x18\x06 \x01(\tR\ntimeColumn\x12\x44\n\rstream_config\x18\x07 \x01(\x0b\x32\x1f.tecton_proto.args.StreamConfigR\x0cstreamConfig\x12\x44\n\x07options\x18\x08 \x03(\x0b\x32*.tecton_proto.data.StreamDataSource.OptionR\x07options\x1a\x30\n\x06Option\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05valueB\x0f\n\rstream_sourceJ\x04\x08\x02\x10\x03J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06\"\x9d\x02\n\x14PushDataSourceConfig\x12\x1f\n\x0blog_offline\x18\x01 \x01(\x08R\nlogOffline\x12M\n\x0epost_processor\x18\x02 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\rpostProcessor\x12>\n\x0cinput_schema\x18\x03 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x0binputSchema\x12U\n\x13post_processor_mode\x18\x04 \x01(\x0e\x32%.tecton_proto.args.TransformationModeR\x11postProcessorModeB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.stream_data_source__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _KINESISDSCONFIG._serialized_start=318
  _KINESISDSCONFIG._serialized_end=410
  _KAFKADSCONFIG._serialized_start=413
  _KAFKADSCONFIG._serialized_end=794
  _SPARKSTREAMDATASOURCEFUNCTION._serialized_start=796
  _SPARKSTREAMDATASOURCEFUNCTION._serialized_end=895
  _STREAMDATASOURCE._serialized_start=898
  _STREAMDATASOURCE._serialized_end=1752
  _STREAMDATASOURCE_OPTION._serialized_start=1669
  _STREAMDATASOURCE_OPTION._serialized_end=1717
  _PUSHDATASOURCECONFIG._serialized_start=1755
  _PUSHDATASOURCECONFIG._serialized_end=2040
# @@protoc_insertion_point(module_scope)



import unittest

import pandas

from tecton.framework.data_frame import TectonDataFrame


class TestDataFrame(unittest.TestCase):
    def test_to_utc_aware_pandas_with_naive_timestamps(self):
        data = {"timestamp": ["2024-01-01 00:00:00", "2024-01-01 12:00:00"]}
        df = pandas.DataFrame(data)
        df["timestamp"] = pandas.to_datetime(df["timestamp"])

        result = TectonDataFrame._cast_timestamps_to_utc_aware_pandas(df)
        expected_data = {"timestamp": ["2024-01-01 00:00:00", "2024-01-01 12:00:00"]}
        expected = pandas.DataFrame(expected_data).astype(
            {
                "timestamp": "datetime64[us]",
            }
        )
        expected["timestamp"] = expected["timestamp"].dt.tz_localize("UTC")

        pandas.testing.assert_frame_equal(result, expected)

    def test_to_utc_aware_pandas_with_non_utc_timezone(self):
        data = {"timestamp": ["2024-01-01 05:00:00+05:00", "2024-01-01 17:00:00+05:00"]}
        df = pandas.DataFrame(data)
        df["timestamp"] = pandas.to_datetime(df["timestamp"]).astype(
            {
                "timestamp": "datetime64[us, America/New_York]",
            }
        )

        result = TectonDataFrame._cast_timestamps_to_utc_aware_pandas(df)
        expected_data = {"timestamp": ["2024-01-01 00:00:00", "2024-01-01 12:00:00"]}
        expected = pandas.DataFrame(expected_data).astype(
            {
                "timestamp": "datetime64[us]",
            }
        )
        expected["timestamp"] = expected["timestamp"].dt.tz_localize("UTC")
        pandas.testing.assert_frame_equal(result, expected)

    def test_to_utc_aware_pandas_with_mixed_timestamps(self):
        data = {
            "timestamp_naive": ["2024-01-01 00:00:00", "2024-01-01 12:00:00"],
            "timestamp_aware": ["2024-01-01 05:00:00+05:00", "2024-01-01 17:00:00+05:00"],
        }
        df = pandas.DataFrame(data)
        df["timestamp_naive"] = pandas.to_datetime(df["timestamp_naive"])
        df["timestamp_aware"] = pandas.to_datetime(df["timestamp_aware"])

        result = TectonDataFrame._cast_timestamps_to_utc_aware_pandas(df)
        expected_data = {
            "timestamp_naive": ["2024-01-01 00:00:00", "2024-01-01 12:00:00"],
            "timestamp_aware": ["2024-01-01 00:00:00", "2024-01-01 12:00:00"],
        }
        expected = pandas.DataFrame(expected_data).astype(
            {
                "timestamp_naive": "datetime64[us]",
                "timestamp_aware": "datetime64[us]",
            }
        )
        expected["timestamp_naive"] = expected["timestamp_naive"].dt.tz_localize("UTC")
        expected["timestamp_aware"] = expected["timestamp_aware"].dt.tz_localize("UTC")

        pandas.testing.assert_frame_equal(result, expected)

    def test_to_utc_aware_pandas_with_already_utc(self):
        data = {"timestamp": ["2024-01-01 00:00:00+00:00", "2024-01-01 12:00:00+00:00"]}
        df = pandas.DataFrame(data)
        df["timestamp"] = pandas.to_datetime(df["timestamp"]).astype(
            {
                "timestamp": "datetime64[us, UTC]",
            }
        )

        result = TectonDataFrame._cast_timestamps_to_utc_aware_pandas(df)
        expected_data = {"timestamp": ["2024-01-01 00:00:00", "2024-01-01 12:00:00"]}
        expected = pandas.DataFrame(expected_data).astype(
            {
                "timestamp": "datetime64[us]",
            }
        )
        expected["timestamp"] = expected["timestamp"].dt.tz_localize("UTC")

        pandas.testing.assert_frame_equal(result, expected)



import attrs
import pandas

from tecton_athena.athena_session import AthenaSession
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_utils import tree_contains
from tecton_core.query.nodes import MultiOdfvPipelineNode
from tecton_core.query.nodes import RenameColsNode
from tecton_core.query.pandas.node import PandasExecNode
from tecton_core.query.pandas.node import SqlExecNode
from tecton_core.query.pandas.nodes import PandasMultiOdfvPipelineNode
from tecton_core.query.pandas.nodes import PandasRenameColsNode
from tecton_core.query.pandas.sql import SqlExecutor


@attrs.frozen
class AthenaSqlExecutor(SqlExecutor):
    session: AthenaSession

    def read_sql(self, sql: str) -> pandas.DataFrame:
        pandas_df = self.session.read_sql(sql)
        return pandas_df


# Converts a logical querytree to a physical Athena querytree by converting the RenameColsNodes and
# MultiOdfvPipelineNode at the top of the tree to PandasExecNodes and the input node directly below these nodes to an
# Athena SqlExecNode that will execute the SQL in an Athena session. The requirement here is that ODFVPipelineNodes will
# not appear in a middle of a querytree and will only ever appear at the top of the querytree.
def athena_convert(node_ref: NodeRef, sql_executor: AthenaSqlExecutor, pretty_sql: bool) -> PandasExecNode:
    if tree_contains(node_ref, MultiOdfvPipelineNode):
        return _convert_to_pandas_nodes(node_ref, sql_executor, pretty_sql)
    else:
        return SqlExecNode.from_sql_inputs(node_ref.node, sql_executor, pretty_sql)


def _convert_to_pandas_nodes(tree: NodeRef, sql_executor, pretty_sql: bool):
    # Recurses over RenameColsNodes and MultiOdfvPipelineNode at the top of the tree and converts them to
    # PandasExecNodes and converts only the node immediately below these nodes to a SqlExecNode
    logical_tree_node = tree.node
    node_mapping = {
        MultiOdfvPipelineNode: PandasMultiOdfvPipelineNode,
        RenameColsNode: PandasRenameColsNode,
    }
    if logical_tree_node.__class__ in node_mapping:
        input_node = _convert_to_pandas_nodes(logical_tree_node.input_node, sql_executor, pretty_sql)
        return node_mapping[logical_tree_node.__class__].from_node_inputs(logical_tree_node, input_node)
    else:
        return SqlExecNode.from_sql_inputs(logical_tree_node, sql_executor, pretty_sql)



from typing import Dict
from typing import List
from typing import Tuple

import attrs

from tecton_core import data_types as tecton_types
from tecton_core.data_types import DataType
from tecton_core.data_types import data_type_from_proto
from tecton_core.errors import TectonValidationError
from tecton_proto.common import schema__client_pb2 as schema_pb2


@attrs.frozen
class Schema:
    proto: schema_pb2.Schema

    def __repr__(self) -> str:
        # NOTE: this __repr__ affects what is printed for spec parity tests.
        # If you are debugging a test failure that you suspect its schema diff related,
        # we recommend commenting out this override so that the full proto is printed.
        return str(self.column_name_and_data_types())

    # TODO(jake): Remove this method. Just access proto attr directly.
    def to_proto(self) -> schema_pb2.Schema:
        return self.proto

    def column_names(self) -> List[str]:
        return [c.name for c in self.proto.columns]

    def column_name_and_data_types(self) -> List[Tuple[str, DataType]]:
        return [(c.name, data_type_from_proto(c.offline_data_type)) for c in self.proto.columns]

    def to_dict(self) -> Dict[str, DataType]:
        return dict(self.column_name_and_data_types())

    def is_equivalent(self, other_schema: "Schema") -> bool:
        """Checks if this schema is equivalent to another schema, ignoring column order.

        Only compares offline data types. Other fields are ignored.
        """
        return self.to_dict() == other_schema.to_dict()

    @classmethod
    def from_dict(cls, schema_dict: Dict[str, DataType]) -> "Schema":
        schema_proto = schema_pb2.Schema()
        for col_name, col_type in schema_dict.items():
            col = schema_proto.columns.add()
            col.CopyFrom(column_from_tecton_data_type(col_type))
            col.name = col_name
        return cls(proto=schema_proto)  # type: ignore

    def __add__(self, other: "Schema") -> "Schema":
        if other is None:
            return self
        self_dict = self.to_dict()
        other_dict = other.to_dict()

        # Check if two schema share any column name with different types. This should never happen theoretically.
        shared_col_name = set(self_dict.keys()).intersection(set(other_dict.keys()))
        for col_name in shared_col_name:
            if self_dict[col_name] != other_dict[col_name]:
                msg = f"Column name '{col_name} has different data types: {str(self_dict[col_name])} VS {str(other_dict[col_name])}"
                raise TectonValidationError(msg)

        self_dict.update(other_dict)
        return self.from_dict(self_dict)


@attrs.frozen
class Column:
    name: str
    dtype: DataType

    @classmethod
    def from_proto(cls, proto: schema_pb2.Column) -> "Column":
        return cls(name=proto.name, dtype=tecton_types.data_type_from_proto(proto.offline_data_type))


def get_feature_server_data_type(offline_data_type: tecton_types.DataType) -> tecton_types.DataType:
    if offline_data_type == tecton_types.Int32Type():
        return tecton_types.Int64Type()
    return offline_data_type


def column_from_tecton_data_type(offline_data_type: tecton_types.DataType) -> schema_pb2.Column:
    return schema_pb2.Column(
        offline_data_type=offline_data_type.proto,
        feature_server_data_type=get_feature_server_data_type(offline_data_type).proto,
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/auth/service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'tecton_proto/auth/service__client.proto\x12\x11tecton_proto.auth\x1a google/protobuf/descriptor.proto\"\xbe\x03\n\x0c\x41uthMetadata\x12\x36\n\x13skip_authentication\x18\x01 \x01(\x08:\x05\x66\x61lseR\x12skipAuthentication\x12\x34\n\x12skip_authorization\x18\x02 \x01(\x08:\x05\x66\x61lseR\x11skipAuthorization\x12.\n\npermission\x18\x05 \x01(\t:\x0enot_configuredR\npermission\x12i\n\x1d\x61\x64vanced_permission_overrides\x18\x06 \x03(\x0b\x32%.tecton_proto.auth.PermissionOverrideR\x1b\x61\x64vancedPermissionOverrides\x12S\n\x12resource_reference\x18\x03 \x01(\x0b\x32$.tecton_proto.auth.ResourceReferenceR\x11resourceReference\x12J\n\x1e\x64\x65\x66\x65r_authorization_to_service\x18\x08 \x01(\x08:\x05\x66\x61lseR\x1b\x64\x65\x66\x65rAuthorizationToServiceJ\x04\x08\x07\x10\x08\"\xa0\x01\n\x12PermissionOverride\x12\x30\n\x14\x63ondition_field_path\x18\x01 \x01(\tR\x12\x63onditionFieldPath\x12\'\n\x0f\x63ondition_value\x18\x02 \x01(\tR\x0e\x63onditionValue\x12/\n\x13permission_override\x18\x03 \x01(\tR\x12permissionOverride\"c\n\x11ResourceReference\x12:\n\x04type\x18\x01 \x01(\x0e\x32&.tecton_proto.auth.ResourceRefTypeEnumR\x04type\x12\x12\n\x04path\x18\x02 \x01(\tR\x04path*\xd6\x01\n\x13ResourceRefTypeEnum\x12!\n\x1dRESOURCE_REF_TYPE_UNSPECIFIED\x10\x00\x12$\n RESOURCE_REF_TYPE_WORKSPACE_NAME\x10\x01\x12(\n$RESOURCE_REF_TYPE_SERVICE_ACCOUNT_ID\x10\x02\x12(\n$RESOURCE_REF_TYPE_PRINCIPAL_GROUP_ID\x10\x03\x12\"\n\x1eRESOURCE_REF_TYPE_SECRET_SCOPE\x10\x04:g\n\rauth_metadata\x12\x1e.google.protobuf.MethodOptions\x18\xc4\xe7\x8cX \x01(\x0b\x32\x1f.tecton_proto.auth.AuthMetadataR\x0c\x61uthMetadataB(Z&github.com/tecton-ai/tecton_proto/auth')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.auth.service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:
  google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(auth_metadata)

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z&github.com/tecton-ai/tecton_proto/auth'
  _RESOURCEREFTYPEENUM._serialized_start=810
  _RESOURCEREFTYPEENUM._serialized_end=1024
  _AUTHMETADATA._serialized_start=97
  _AUTHMETADATA._serialized_end=543
  _PERMISSIONOVERRIDE._serialized_start=546
  _PERMISSIONOVERRIDE._serialized_end=706
  _RESOURCEREFERENCE._serialized_start=708
  _RESOURCEREFERENCE._serialized_end=807
# @@protoc_insertion_point(module_scope)



import random
import string
from datetime import datetime
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas
import pendulum
from pyspark.sql.types import StructType

from tecton_core import specs
from tecton_core.errors import UDF_ERROR
from tecton_core.errors import UDF_TYPE_ERROR
from tecton_core.id_helper import IdHelper
from tecton_core.materialization_context import BoundMaterializationContext
from tecton_core.materialization_context import MaterializationContext
from tecton_core.pandas_compat import pandas_to_spark
from tecton_core.pipeline_common import CONSTANT_TYPE
from tecton_core.pipeline_common import check_transformation_type
from tecton_core.pipeline_common import constant_node_to_value
from tecton_core.pipeline_common import get_keyword_inputs
from tecton_core.pipeline_common import positional_inputs
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.args.pipeline__client_pb2 import TransformationNode
from tecton_proto.args.transformation__client_pb2 import TransformationMode
from tecton_proto.common.id__client_pb2 import Id as TectonId
from tecton_spark.type_annotations import PySparkDataFrame
from tecton_spark.type_annotations import PySparkSession


class SparkFeaturePipeline:
    # Each internal node in the Pipeline Tree evaluates to a single type of value which can
    # be one of the following:
    _VALUE_TYPE = Union[PySparkDataFrame, CONSTANT_TYPE, MaterializationContext]

    # note that pipeline is included since this is meant to be a user hint, and it's
    # theoretically possible that a pipeline mode node shows up deeper than expected
    _VALID_MODES = ["pyspark", "spark_sql", "pipeline", "python", "pandas"]

    def __init__(
        self,
        spark: PySparkSession,
        pipeline: PipelineNode,
        transformations: List[specs.TransformationSpec],
        materialization_context_limits: Optional[pendulum.Period] = None,
        schedule_interval: Optional[pendulum.Duration] = None,
        data_source_inputs: Optional[Dict[str, PySparkDataFrame]] = None,
        # output_schema is only used by python/pandas transformations during backfills.
        # Specifically, it applies for Stream feature views with push sources and batch config, during
        # the batch materialization jobs.
        output_schema: Optional[StructType] = None,
    ) -> None:
        self._spark = spark
        self._pipeline = pipeline
        self._output_schema = output_schema
        self._materialization_context_limits = materialization_context_limits
        self._id_to_transformation = {t.id: t for t in transformations}
        self._schedule_interval = schedule_interval
        self._data_source_inputs = data_source_inputs
        self._registered_temp_view_names: List[str] = []

    def get_transformation_by_id(self, id: TectonId) -> specs.TransformationSpec:
        return self._id_to_transformation[IdHelper.to_string(id)]

    def to_dataframe(self):
        return self.get_dataframe()

    def get_dataframe(self) -> PySparkDataFrame:
        df = self._node_to_value(self._pipeline.root)
        self._cleanup_temp_tables()
        assert isinstance(df, PySparkDataFrame)
        return df

    def _node_to_value(self, pipeline_node: PipelineNode) -> _VALUE_TYPE:
        if pipeline_node.HasField("transformation_node"):
            return self._transformation_node_to_value(pipeline_node.transformation_node)
        elif pipeline_node.HasField("data_source_node"):
            data_source_node = pipeline_node.data_source_node
            if data_source_node.input_name not in self._data_source_inputs:
                msg = f"Expected inputs {self._data_source_inputs} to contain {data_source_node.input_name}"
                raise ValueError(msg)
            return self._data_source_inputs[data_source_node.input_name]
        elif pipeline_node.HasField("constant_node"):
            return constant_node_to_value(pipeline_node.constant_node)
        elif pipeline_node.HasField("materialization_context_node") or pipeline_node.HasField("context_node"):
            if self._materialization_context_limits is not None:
                feature_start_time = self._materialization_context_limits.start
                feature_end_time = self._materialization_context_limits.end
                batch_schedule = self._schedule_interval
            else:
                feature_start_time = pendulum.from_timestamp(0, pendulum.tz.UTC)
                feature_end_time = pendulum.datetime(2100, 1, 1)
                batch_schedule = self._schedule_interval or pendulum.duration()
            return BoundMaterializationContext._create_internal(feature_start_time, feature_end_time, batch_schedule)
        elif pipeline_node.HasField("request_data_source_node"):
            msg = "RequestDataSource is not supported in Spark pipelines"
            raise ValueError(msg)
        elif pipeline_node.HasField("feature_view_node"):
            msg = "Dependent FeatureViews are not supported in Spark pipelines"
            raise ValueError(msg)
        else:
            msg = f"Unknown PipelineNode type: {pipeline_node}"
            raise KeyError(msg)

    def _transformation_node_to_value(self, transformation_node: TransformationNode) -> PySparkDataFrame:
        """Recursively translates inputs to values and then passes them to the transformation."""
        args: List[Union[PySparkDataFrame, str, int, float, bool, datetime]] = []
        kwargs = {}
        for transformation_input in transformation_node.inputs:
            node_value = self._node_to_value(transformation_input.node)
            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                msg = f"Unknown argument type for Input node: {transformation_input}"
                raise KeyError(msg)

        return self._apply_transformation_function(transformation_node, args, kwargs)

    def _apply_transformation_function(
        self, transformation_node: TransformationNode, args: List[Any], kwargs: Dict[str, Any]
    ) -> Union[Dict[str, Any], pandas.DataFrame, PySparkDataFrame]:
        """For the given transformation node, returns the corresponding DataFrame transformation.

        If needed, resulted function is wrapped with a function that translates mode-specific input/output types to DataFrames.

        Pandas and Python mode cases below run for IngestAPI. See `RealtimeFeaturePipeline` class
        for Realtime Feature Views.
        """
        transformation = self.get_transformation_by_id(transformation_node.transformation_id)
        mode = transformation.transformation_mode
        user_function = transformation.user_function

        if mode == TransformationMode.TRANSFORMATION_MODE_PYSPARK:
            try:
                res = user_function(*args, **kwargs)
            except Exception as e:
                raise UDF_ERROR(e, feature_definition_name=transformation.metadata.name)
            check_transformation_type(transformation.name, res, "pyspark", self._VALID_MODES)
            return res
        elif mode == TransformationMode.TRANSFORMATION_MODE_SPARK_SQL:
            return self._build_spark_sql_function(transformation_node, user_function)(*args, **kwargs)
        elif mode == TransformationMode.TRANSFORMATION_MODE_PANDAS:
            try:
                return self._run_pandas_function(user_function, args, kwargs)
            except Exception as e:
                raise UDF_ERROR(e, feature_definition_name=transformation.metadata.name)
        elif mode == TransformationMode.TRANSFORMATION_MODE_PYTHON:
            try:
                return self._create_and_run_python_udf(user_function, args, kwargs)
            except TypeError as e:
                raise UDF_TYPE_ERROR(e)
            except Exception as e:
                raise UDF_ERROR(e, feature_definition_name=transformation.metadata.name)
        else:
            msg = f"Unknown transformation mode({transformation.transformation_mode})"
            raise KeyError(msg)

    def _run_pandas_function(
        self,
        user_function: Callable,
        args: List[Any],
        kwargs: Dict[str, Any],
    ) -> PySparkDataFrame:
        # Assumes that we only have one argument that is a pyspark Dataframe built from the batch source
        # because this code path is for ingest api which should only read from one batch source.
        assert len(args) + len(kwargs) == 1, "Pandas transformations only support a single input"
        input_df = args[0] if len(args) == 1 else next(iter(kwargs.values()))

        output_pandas_df = user_function(input_df.toPandas())
        return pandas_to_spark(self._spark, output_pandas_df)

    def _create_and_run_python_udf(
        self, user_function: Callable, args: List[Any], kwargs: Dict[str, Any]
    ) -> pandas.DataFrame:
        from pyspark.sql.functions import struct
        from pyspark.sql.functions import udf

        # Assumes that we only have one argument that is a pyspark Dataframe built from the batch source
        # because this code path is for ingest api which should only read from one batch source.
        assert len(args) + len(kwargs) == 1, "Pandas transformations only support a single input"
        df = args[0] if len(args) == 1 else next(iter(kwargs.values()))

        @udf(self._output_schema)
        def transform_rows(row_as_struct):
            # Apply your desired transformation on the input group of Rows(this applies row by row)
            transformed_df_group = user_function(row_as_struct.asDict())

            # Return the transformed group of rows as a Pandas DataFrame
            return transformed_df_group

        df = df.select(struct("*").alias("data"))
        df = df.select(transform_rows("data").alias("result"))
        res = df.select(*[f"result.{field.name}" for field in self._output_schema])
        return res

    def _build_spark_sql_function(
        self, transformation_node: TransformationNode, user_function: Callable[..., str]
    ) -> Callable[..., PySparkDataFrame]:
        """
        Returns a function that takes args/kwargs and appropriately passes it as arguments
        to a formatted SQL string.
        """

        def wrapped(*args, **kwargs):
            wrapped_args = []
            for input_value, input_proto in zip(args, positional_inputs(transformation_node)):
                # Replace Dataframe arguments with temp table name
                if isinstance(input_value, PySparkDataFrame):
                    input_value = self._register_temp_table(input_proto.node, input_value)
                wrapped_args.append(input_value)

            keyword_inputs = get_keyword_inputs(transformation_node)
            wrapped_kwargs = {}
            for k, input_value in kwargs.items():
                input_proto = keyword_inputs[k]
                # Replace Dataframe arguments with temp table name
                if isinstance(input_value, PySparkDataFrame):
                    input_value = self._register_temp_table(input_proto.node, input_value)
                wrapped_kwargs[k] = input_value

            sql_string = user_function(*wrapped_args, **wrapped_kwargs)

            transformation_name = self.get_transformation_by_id(transformation_node.transformation_id).name
            check_transformation_type(transformation_name, sql_string, "spark_sql", self._VALID_MODES)
            return self._spark.sql(sql_string)

        return wrapped

    def _register_temp_table(self, node: PipelineNode, df: PySparkDataFrame) -> str:
        """
        Registers a Dataframe as a temp table and returns its generated name.
        Only applicable for Transformation and DataSource Nodes
        """
        if node.HasField("transformation_node"):
            transformation_name = self.get_transformation_by_id(node.transformation_node.transformation_id).name
            name = f"transformation_{transformation_name}_output"
        elif node.HasField("data_source_node"):
            name = node.data_source_node.input_name
        else:
            msg = f"Expected transformation or data source node: {node}"
            raise Exception(msg)

        random_suffix = "".join(random.choice(string.ascii_letters) for i in range(6))
        unique_name = name + random_suffix

        self._registered_temp_view_names.append(unique_name)
        df.createOrReplaceTempView(unique_name)

        return unique_name

    def _cleanup_temp_tables(self):
        # Cleanup any temporary tables created during the process
        for temp_name in self._registered_temp_view_names:
            # DROP VIEW/DROP TABLE sql syntax is invalidated when spark_catalog is set to DeltaCatalog on EMR clusters
            if (
                self._spark.conf.get("spark.sql.catalog.spark_catalog", "")
                == "org.apache.spark.sql.delta.catalog.DeltaCatalog"
            ):
                self._spark.catalog.dropTempView(temp_name)
            else:
                self._spark.sql(f"DROP VIEW IF EXISTS {temp_name}")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/time_window__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n-tecton_proto/common/time_window__client.proto\x12\x13tecton_proto.common\x1a\x1egoogle/protobuf/duration.proto\"\x9a\x02\n\nTimeWindow\x12[\n\x14relative_time_window\x18\x01 \x01(\x0b\x32\'.tecton_proto.common.RelativeTimeWindowH\x00R\x12relativeTimeWindow\x12N\n\x0flifetime_window\x18\x02 \x01(\x0b\x32#.tecton_proto.common.LifetimeWindowH\x00R\x0elifetimeWindow\x12U\n\x12time_window_series\x18\x03 \x01(\x0b\x32%.tecton_proto.common.TimeWindowSeriesH\x00R\x10timeWindowSeriesB\x08\n\x06window\"^\n\x10TimeWindowSeries\x12J\n\x0ctime_windows\x18\x01 \x03(\x0b\x32\'.tecton_proto.common.RelativeTimeWindowR\x0btimeWindows\"\x8c\x01\n\x12RelativeTimeWindow\x12<\n\x0cwindow_start\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0bwindowStart\x12\x38\n\nwindow_end\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\twindowEnd\"\x10\n\x0eLifetimeWindowB\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.time_window__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _TIMEWINDOW._serialized_start=103
  _TIMEWINDOW._serialized_end=385
  _TIMEWINDOWSERIES._serialized_start=387
  _TIMEWINDOWSERIES._serialized_end=481
  _RELATIVETIMEWINDOW._serialized_start=484
  _RELATIVETIMEWINDOW._serialized_end=624
  _LIFETIMEWINDOW._serialized_start=626
  _LIFETIMEWINDOW._serialized_end=642
# @@protoc_insertion_point(module_scope)



import logging
import operator
from datetime import timedelta
from functools import reduce
from typing import ClassVar
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import attrs
import pendulum
import pyspark
import pyspark.sql.functions as F
import pyspark.sql.window as spark_window

from tecton_core import time_utils
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query import compaction_utils
from tecton_core.query_consts import aggregation_group_id
from tecton_core.query_consts import anchor_time
from tecton_core.query_consts import tecton_secondary_key_aggregation_indicator_col
from tecton_core.query_consts import temp_indictor_column_name
from tecton_core.query_consts import temp_struct_column_name
from tecton_core.schema import Schema
from tecton_core.specs.feature_view_spec import OnlineBatchTablePart
from tecton_core.specs.time_window_spec import LifetimeWindowSpec
from tecton_core.specs.time_window_spec import RelativeTimeWindowSpec
from tecton_core.specs.time_window_spec import TimeWindowSeriesSpec
from tecton_core.specs.time_window_spec import TimeWindowSpec
from tecton_core.specs.time_window_spec import create_time_window_spec_from_data_proto
from tecton_core.time_utils import convert_timedelta_for_version
from tecton_core.time_utils import convert_timestamp_for_version
from tecton_proto.data import feature_view__client_pb2 as feature_view_pb2
from tecton_spark.aggregation_plans import get_aggregation_plan
from tecton_spark.query.node import SparkExecNode


logger = logging.getLogger(__name__)


@attrs.frozen
class JoinSparkNode(SparkExecNode):
    """
    A basic left join on 2 inputs
    """

    left: SparkExecNode
    right: SparkExecNode
    join_cols: List[str]
    how: str
    allow_nulls: bool = False

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        # NOTE: the aliases are added because on Databricks the usage of this node in a
        # FeatureService join errors when going through the allow nulls codepath.
        # The error claimed an ambiguous self-join, when the spine events are the same data source as a feature view's events.
        left_df = self.left.to_dataframe(spark).alias("left_join_df")
        right_df = self.right.to_dataframe(spark).alias("right_join_df")
        if self.allow_nulls:
            join_condition = reduce(operator.and_, [left_df[col].eqNullSafe(right_df[col]) for col in self.join_cols])
            right_nonjoin_cols = set(self.right.columns) - set(self.join_cols)
            # We need to quote the column names because we use dots in the column names to separate feature view name and feature name.
            return left_df.join(right_df, how=self.how, on=join_condition).select(
                *[left_df[f"`{col}`"] for col in self.left.columns],
                *[right_df[f"`{col}`"] for col in right_nonjoin_cols],
            )
        else:
            return left_df.join(right_df, how=self.how, on=self.join_cols)


@attrs.frozen
class WildcardJoinSparkNode(SparkExecNode):
    left: SparkExecNode
    right: SparkExecNode
    join_cols: List[str]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        left_df = self.left.to_dataframe(spark)
        right_df = self.right.to_dataframe(spark)

        left_columns = [getattr(left_df, col) for col in left_df.columns if col not in self.join_cols]
        right_columns = [getattr(right_df, col) for col in right_df.columns if col not in self.join_cols]

        left_df = left_df.withColumn("_tecton_join_cols", F.struct(self.join_cols))
        right_df = right_df.withColumn("_tecton_join_cols", F.struct(self.join_cols))

        return left_df.join(right_df, how="outer", on=["_tecton_join_cols"]).select(
            "_tecton_join_cols.*", *left_columns, *right_columns
        )


@attrs.frozen
class AsofJoinInputSparkContainer:
    node: SparkExecNode
    timestamp_field: str  # spine or feature timestamp
    effective_timestamp_field: Optional[str]
    prefix: Optional[str]
    schema: Optional[Schema]


@attrs.frozen
class _AsofJoinerAndWindowAggregator:
    """
    Implements an asof join + window aggregate, with the left side being the
    'spine' and the right side being used for the aggregate value.

    There are a few ways this behavior can be implemented, but by test the best
    performing method has been to union the two inputs and use a window
    function over the unioned DataFrame to do the calculation.

    This can be used for a traditional asof join by using a `last` window func on the unioned dataframe:
        last(col) OVER (PARTITION BY partition_cols ORDER BY timestamp_cols RANGE BETWEEN unbounded preceding AND current row)

    This is also used by the full aggregate rollup (which uses more complicated set of window specs and rollup functions). Example 1h count rollup:
        sum(count_col) OVER (PARTITION BY partition_cols ORDER BY timestamp_cols RANGE BETWEEN 3600 AND current row)
    """

    _left_df: pyspark.sql.DataFrame
    _right_df: pyspark.sql.DataFrame
    timestamp_cols: Tuple[str, ...]
    common_partition_cols: Tuple[str, ...]  # partition columns used by all aggregations.
    _extra_partition_cols: Tuple[str, ...]  # partition columns not shared by every aggregation.
    _left_ts_cols: Tuple[str, ...]
    _use_window_range_between_value: bool
    _tecton_window_range_between_order_columns: Tuple[str, ...]

    _timestamp_prefix: ClassVar[str] = "_tecton_asof_join_timestamp"
    _left_prefix: ClassVar[str] = "_tecton_left"
    _is_left_column_name: ClassVar[str] = "IS_LEFT"
    _tecton_window_range_between_order_col_name: ClassVar[str] = "_tecton_window_range_between_order_col"

    @classmethod
    def create(
        cls,
        left_df: pyspark.sql.DataFrame,
        right_df: pyspark.sql.DataFrame,
        left_ts_cols: Sequence[str],
        right_ts_cols: Sequence[str],
        partition_cols: Sequence[str],
        use_window_range_between_value: bool,
        extra_partition_cols: Sequence[str] = [],
    ) -> "_AsofJoinerAndWindowAggregator":
        if len(left_ts_cols) != len(right_ts_cols):
            msg = f"Timestamp columns are not equal length: left({left_ts_cols}), right({right_ts_cols})"
            raise RuntimeError(msg)

        timestamp_cols = [f"{cls._timestamp_prefix}_{i}" for i in range(len(left_ts_cols))]
        window_range_between_order_cols = [
            f"{cls._tecton_window_range_between_order_col_name}_{i}" for i in range(len(timestamp_cols))
        ]

        left_df_cols = [F.col(old_col).alias(new_col) for new_col, old_col in zip(timestamp_cols, left_ts_cols)]
        left_df = left_df.select("*", *left_df_cols)

        right_df_cols = [F.col(old_col).alias(new_col) for new_col, old_col in zip(timestamp_cols, right_ts_cols)]
        right_df = right_df.select("*", *right_df_cols)

        return cls(
            left_df=left_df,
            right_df=right_df,
            timestamp_cols=tuple(timestamp_cols),
            left_ts_cols=tuple(left_ts_cols),
            common_partition_cols=tuple(partition_cols),
            extra_partition_cols=tuple(extra_partition_cols),
            use_window_range_between_value=use_window_range_between_value,
            tecton_window_range_between_order_columns=tuple(window_range_between_order_cols),
        )

    @property
    def common_cols(self) -> List[str]:
        return list(self.timestamp_cols) + list(self.all_partition_columns)

    @property
    def all_partition_columns(self) -> List[str]:
        return list(self.common_partition_cols) + list(self._extra_partition_cols)

    @property
    def left_nonjoin_cols(self) -> List[str]:
        return [c for c in self._left_df.columns if c not in set(self.common_cols)]

    @property
    def _right_nonjoin_cols(self) -> List[str]:
        return [c for c in self._right_df.columns if c not in set(self.common_cols)]

    def _union(self) -> pyspark.sql.DataFrame:
        # schemas have to match exactly so that the 2 dataframes can be unioned together.
        left_full_cols = (
            [F.lit(True).alias(self._is_left_column_name)]
            + [F.col(x) for x in self.common_cols]
            + [F.col(x).alias(f"{self._left_prefix}_{x}") for x in self.left_nonjoin_cols]
            + [F.lit(None).alias(x) for x in self._right_nonjoin_cols]
        )
        right_full_cols = (
            [F.lit(False).alias(self._is_left_column_name)]
            + [F.col(x) for x in self.common_cols]
            + [F.lit(None).alias(f"{self._left_prefix}_{x}") for x in self.left_nonjoin_cols]
            + [F.col(x) for x in self._right_nonjoin_cols]
        )

        if self._use_window_range_between_value:
            # we want left rows to be sorted after right rows if timestamps are the same. But when using "window range between"
            # only 1 column can be used for sorting. As a workaround, we multiply the left rows by * 2 and then add 1. The right
            # rows will be multiplied by 2 only. This will ensure the left rows will be sorted after the right rows.
            for index, timestamp_col in enumerate(self.timestamp_cols):
                order_by_col = self._tecton_window_range_between_order_columns[index]
                left_full_cols.append((F.col(timestamp_col).cast("long") * 2 + 1).alias(order_by_col))
                right_full_cols.append((F.col(timestamp_col).cast("long") * 2).alias(order_by_col))

        left_df = self._left_df.select(left_full_cols)
        right_df = self._right_df.select(right_full_cols)
        return left_df.union(right_df)

    def join_and_aggregate(self, aggregations: List[pyspark.sql.Column]) -> pyspark.sql.DataFrame:
        union = self._union()

        # We use the right side of asof join to calculate the aggregate values to augment to the rows from the left side.
        # Then, we drop the right side's rows.
        output_columns = (
            self.common_cols
            + [F.col(f"{self._left_prefix}_{x}").alias(x) for x in self.left_nonjoin_cols]
            + aggregations
            + [self._is_left_column_name]
        )
        selected = union.select(output_columns)
        return selected.filter(self._is_left_column_name).drop(self._is_left_column_name, *self.timestamp_cols)

    def get_range_between_window_spec(
        self,
        range_between_start: spark_window.Window,
        range_between_end: spark_window.Window,
        timestamp_col: Optional[str] = None,
        extra_partition_cols: List[str] = [],
    ) -> spark_window.WindowSpec:
        assert len(extra_partition_cols) == 0 or all(
            col in self._extra_partition_cols for col in extra_partition_cols
        ), f"Expected extra_partition_cols {extra_partition_cols} in {self._extra_partition_cols}."
        assert (
            timestamp_col is None or timestamp_col in self._left_ts_cols
        ), f"Expected timestamp_col {timestamp_col} in {self._left_ts_cols}."

        if range_between_start == spark_window.Window.unboundedPreceding:
            order_by_cols = [*list(self.timestamp_cols), self._is_left_column_name]
        else:
            if timestamp_col:
                order_column_index = self._left_ts_cols.index(timestamp_col)
            else:
                order_column_index = 0
            order_by_col = self._tecton_window_range_between_order_columns[order_column_index]
            order_by_cols = [order_by_col]

        window_spec = (
            spark_window.Window.partitionBy(list(self.common_partition_cols) + extra_partition_cols)
            .orderBy([F.col(c).asc() for c in order_by_cols])
            .rangeBetween(range_between_start, range_between_end)
        )
        return window_spec


@attrs.frozen
class AsofJoinSparkNode(SparkExecNode):
    """
    A "basic" asof join on 2 inputs.
    LEFT asof_join RIGHT has the following behavior:
        For each row on the left side, find the latest (but <= in time) matching (by join key) row on the right side, and associate the right side's columns to that row.
    The result is a dataframe with the same number of rows as LEFT, with additional columns. These additional columns are prefixed with f"{right_prefix}_". This is the built-in behavior of the tempo library.

    """

    left_container: AsofJoinInputSparkContainer
    right_container: AsofJoinInputSparkContainer
    join_cols: List[str]

    _right_struct_col: ClassVar[str] = "_right_values_struct"

    def _structify_right(self, right_df: pyspark.sql.DataFrame, struct_col_name: str) -> pyspark.sql.DataFrame:
        # we additionally include the right time field though we join on the left's time field.
        # This is so we can see how old the row we joined against is and later determine whether to exclude on basis of ttl
        right_nonjoin_cols = list(set(right_df.columns) - set(self.join_cols))
        # wrap fields on the right in a struct. This is to work around null feature values and ignorenulls
        # used during joining/window function.
        cols_to_wrap = [F.col(c).alias(f"{self.right_container.prefix}_{c}") for c in right_nonjoin_cols]
        right_df = right_df.withColumn(struct_col_name, F.struct(*cols_to_wrap))
        return right_df

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        left_df = self.left_container.node.to_dataframe(spark)
        right_df = self.right_container.node.to_dataframe(spark)

        right_df = self._structify_right(right_df, self._right_struct_col)

        # The left and right dataframes are unioned together and sorted using 2 columns.
        # The spine will use the spine timestamp and the features will be ordered by their
        # (effective_timestamp, feature_timestamp) because multiple features can have the same effective
        # timestamp. We want to return the closest feature to the spine timestamp that also satisfies
        # the condition => effective timestamp <= spine timestamp.
        join_spec = _AsofJoinerAndWindowAggregator.create(
            left_df=left_df,
            right_df=right_df,
            left_ts_cols=[self.left_container.timestamp_field, self.left_container.timestamp_field],
            right_ts_cols=[self.right_container.effective_timestamp_field, self.right_container.timestamp_field],
            partition_cols=self.join_cols,
            use_window_range_between_value=False,
        )
        window_spec = join_spec.get_range_between_window_spec(
            spark_window.Window.unboundedPreceding, spark_window.Window.currentRow
        )
        aggregations = [
            F.last(F.col(self._right_struct_col), ignorenulls=True).over(window_spec).alias(self._right_struct_col)
        ]
        spine_with_features_df = join_spec.join_and_aggregate(aggregations)

        # unwrap the struct to return the fields
        final_df = spine_with_features_df.select(
            self.join_cols + join_spec.left_nonjoin_cols + [f"{self._right_struct_col}.*"]
        )
        return final_df


@attrs.frozen
class AsofJoinFullAggSparkNode(SparkExecNode):
    """
    An asof join very similar to AsofJoinNode, but with a change where it does
    the full aggregation rollup (rather than a last).

    NOTE: This should only be used for window aggregates.

    LEFT asof_join RIGHT has the following behavior:
        For each row in the spine, find the matching partial aggregates (by time range)
        and run the appropriate full aggregate over those rows.

    The result is a dataframe with the same number of rows as the spine, with
    additional columns of the fully aggregated features (or null).

    There are a few ways this behavior can be implemented, but by test the best
    performing method has been to union the two inputs and use a window
    function for the aggregates.
    """

    spine: SparkExecNode
    partial_agg_node: SparkExecNode
    fdw: FeatureDefinitionWrapper
    sawtooth_aggregation_data: Optional[compaction_utils.SawtoothAggregationData] = None
    TECTON_AGG_WINDOW_ORDER_COL: ClassVar[str] = "_tecton_agg_window_order_col"

    def _get_aggregations(self, join_spec: _AsofJoinerAndWindowAggregator) -> List[pyspark.sql.Column]:
        time_aggregation = self.fdw.trailing_time_window_aggregation()
        aggregations = []
        for feature in time_aggregation.features:
            col_datatype = self.fdw.view_schema.to_dict()[feature.input_feature_name]
            aggregation_plan = get_aggregation_plan(
                feature.function,
                feature.function_params,
                time_aggregation.is_continuous,
                time_aggregation.time_key,
                col_datatype,
            )
            names = aggregation_plan.materialized_column_names(feature.input_feature_name)
            time_window = create_time_window_spec_from_data_proto(feature.time_window)
            if isinstance(time_window, TimeWindowSeriesSpec):
                window_aggs = []
                for relative_time_window in time_window.time_windows:
                    window_aggs.append(
                        aggregation_plan.full_aggregation_transform(
                            names,
                            self._generate_aggregation_window_spec(
                                RelativeTimeWindowSpec(
                                    window_start=relative_time_window.window_start,
                                    window_end=relative_time_window.window_end,
                                ),
                                join_spec,
                                feature,
                            ),
                        )
                    )
                agg = F.array(window_aggs)
            else:
                agg = aggregation_plan.full_aggregation_transform(
                    names,
                    self._generate_aggregation_window_spec(
                        time_window,
                        join_spec,
                        feature,
                    ),
                )
            aggregations.append(agg.alias(feature.output_feature_name))

        # If the aggregation secondary key appears, we run a window based aggregation on the secondary key
        # indicator column to identify if the secondary appears in each distinct window. `sum` function is arbitrary
        # here, and we just need a function to return `null` value if the secondary key does not appear in the window.
        if self.fdw.aggregation_secondary_key:
            for secondary_key_output in self.fdw.materialized_fv_spec.secondary_key_rollup_outputs:
                aggregations.append(
                    F.sum(tecton_secondary_key_aggregation_indicator_col())
                    .over(
                        self._generate_aggregation_window_spec(
                            secondary_key_output.time_window,
                            join_spec,
                            feature,
                        )
                    )
                    .alias(
                        f"{tecton_secondary_key_aggregation_indicator_col()}_{secondary_key_output.time_window.to_string()}"
                    )
                )
        return aggregations

    def _generate_aggregation_window_spec(
        self,
        time_window: TimeWindowSpec,
        join_spec: _AsofJoinerAndWindowAggregator,
        feature: feature_view_pb2.AggregateFeature,
    ) -> spark_window.WindowSpec:
        # Since the spine and feature rows are unioned together, the spine rows must be ordered after the feature rows
        # when they have the same timestamp for window aggregation to be correct. Window aggregation does not allow
        # ordering using two columns when range between is used. As a workaround, we multiply the left rows by (* 2 + 1).
        # The right rows by * 2. This will ensure the left rows will be sorted after the right rows.
        # We need to make an adjustment here to earliest_anchor_time due to these changes.
        if isinstance(time_window, RelativeTimeWindowSpec):
            start, end = self.fdw.time_range_for_relative_time_window(time_window)
            if not feature.HasField("batch_sawtooth_tile_size"):
                return join_spec.get_range_between_window_spec(start * 2 - 1, end * 2 - 1)

            sawtooth_tile_size = time_utils.proto_to_duration(feature.batch_sawtooth_tile_size)
            timestamp_col = self.sawtooth_aggregation_data.get_anchor_time_column_for_sawtooth_size(sawtooth_tile_size)
            return join_spec.get_range_between_window_spec(
                start * 2 - 1,
                end * 2 - 1,
                timestamp_col=timestamp_col,
                extra_partition_cols=[
                    self.sawtooth_aggregation_data.get_identifier_partition_column_for_sawtooth_size(sawtooth_tile_size)
                ],
            )
        elif isinstance(time_window, LifetimeWindowSpec):
            return join_spec.get_range_between_window_spec(
                spark_window.Window.unboundedPreceding, spark_window.Window.currentRow
            )
        elif isinstance(time_window, TimeWindowSeriesSpec):
            start, end = self.fdw.time_range_for_relative_time_window(
                RelativeTimeWindowSpec(time_window.window_series_start, time_window.window_series_end)
            )
            return join_spec.get_range_between_window_spec(start * 2 - 1, end * 2 - 1)
        else:
            msg = f"Invalid time_window type: {type(time_window)}"
            raise TypeError(msg)

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        left_df = self.spine.to_dataframe(spark)
        right_df = self.partial_agg_node.to_dataframe(spark)

        partition_columns = self.fdw.join_keys
        agg_specific_partition_columns = (
            self.sawtooth_aggregation_data.get_identifier_partition_columns()
            if self.sawtooth_aggregation_data is not None
            else []
        )

        if self.fdw.aggregation_secondary_key:
            partition_columns.append(self.fdw.aggregation_secondary_key)

            # If the aggregation secondary key appears, we add an indicator column to the partial aggregation result to
            # help identify all secondary keys that are present in each distinct time window.
            right_df = right_df.withColumn(tecton_secondary_key_aggregation_indicator_col(), F.lit(1))

        timestamp_cols = (
            self.sawtooth_aggregation_data.get_anchor_time_columns()
            if self.sawtooth_aggregation_data is not None
            else [anchor_time()]
        )
        join_spec = _AsofJoinerAndWindowAggregator.create(
            left_df=left_df,
            right_df=right_df,
            left_ts_cols=timestamp_cols,
            right_ts_cols=timestamp_cols,
            partition_cols=partition_columns,
            extra_partition_cols=agg_specific_partition_columns,
            use_window_range_between_value=True,
        )
        aggregations = self._get_aggregations(join_spec)

        return join_spec.join_and_aggregate(aggregations)


@attrs.frozen
class AsofSecondaryKeyExplodeSparkNode(SparkExecNode):
    """
    This node explodes a spine that misses the secondary key, into a spine that has the secondary key. This
    explosion is needed in two scenarios:
        1. A Feature View with an aggregation_secondary_key. An aggregation_secondary_key is never in the spine,
        so we always need to explode the spine for it.
        2. A Feature View with a wild card join key. A wild card join key is optional in the spine, so we need to
        explode the spine if and only if the wild card join key is not present.

    This node looks back the max aggregation interval or TTL of the feature view to find the secondary key values by
    using a window based `collect_set` function with an as-of join between left and right dataframes. Using the max
    aggregation interval can help us find all the secondary key values in all windows with a single window based
    `collect_set` function.

    E.g. Let's say a FV has fully bound join_key `A` and a secondary key `C`.
    For every row `[a_0, anchor_0]` from the spine, we will have the following rows in the
    returned dataframe:
       [a_0  c_1  anchor_0]
       [a_0  c_2  anchor_0]
        .    .    .
       [a_0  c_k  anchor_0]
    where (`c_1`, ..., `c_k`) represent all the secondary key values such that, the following row is
    present inside `right`:
        [a_0, c_i, anchor_i]
    and:
        anchor_0 - max_feature_agg_period (or ttl) < anchor_i <= anchor_0.

    Attributes:
         left: The spine node that misses the secondary key.
         left_ts: The timestamp column of the spine node.
         right: The feature value node that contains the secondary key.
         right_ts: The timestamp column of the feature value node.

    """

    left: SparkExecNode
    left_ts: str
    right: SparkExecNode
    right_ts: str
    fdw: FeatureDefinitionWrapper
    TECTON_INTERNAL_SECONDARY_KEY_SET_COL: ClassVar[str] = "_tecton_internal_secondary_key_set"

    @property
    def secondary_key(self) -> str:
        if self.fdw.is_temporal_aggregate:
            return self.fdw.aggregation_secondary_key or self.fdw.wildcard_join_key
        return self.fdw.wildcard_join_key

    @property
    def partition_cols(self) -> List[str]:
        # If the secondary key is an aggregation secondary key, we simply use the join keys as the partition columns.
        # If the secondary key is a wildcard join key, we use all join keys except the wildcard join key as the
        # partition columns.
        if self.fdw.wildcard_join_key:
            return [col for col in self.fdw.join_keys if col != self.fdw.wildcard_join_key]
        elif self.fdw.is_temporal_aggregate and self.fdw.aggregation_secondary_key:
            return self.fdw.join_keys
        else:
            msg = "AsofSecondaryKeyExplodeSparkNode requires either a wildcard join key or an aggregation secondary key"
            raise ValueError(msg)

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        left_df = self.left.to_dataframe(spark)
        right_df = self.right.to_dataframe(spark)

        join_spec = _AsofJoinerAndWindowAggregator.create(
            left_df=left_df,
            right_df=right_df,
            left_ts_cols=[self.left_ts],
            right_ts_cols=[self.right_ts],
            partition_cols=self.partition_cols,
            use_window_range_between_value=True,
        )

        if self.fdw.is_temporal_aggregate and self.fdw.has_lifetime_aggregate:
            earliest_anchor_time = spark_window.Window.unboundedPreceding
        else:
            if self.fdw.is_temporal_aggregate:
                earliest_anchor_time = self.fdw.earliest_anchor_time_from_window_start(self.fdw.earliest_window_start)
            else:
                # Add 1 since we expire at `ttl` time
                max_aggregation_window = int(self.fdw.serving_ttl.total_seconds())
                earliest_anchor_time = -max_aggregation_window + 1

            # Since the spine and feature rows are unioned together, the spine rows must be ordered after the feature rows
            # when they have the same timestamp for window aggregation to be correct. Window aggregation does not allow
            # ordering using two columns when range between is used. As a workaround, we multiply the left rows by (* 2 + 1).
            # The right rows by * 2. This will ensure the left rows will be sorted after the right rows.
            # We need to make an adjustment here to earliest_anchor_time due to these changes.
            earliest_anchor_time = earliest_anchor_time * 2 - 1

        window_spec = join_spec.get_range_between_window_spec(earliest_anchor_time, spark_window.Window.currentRow)
        aggregations = [
            F.collect_set(F.col(self.secondary_key)).over(window_spec).alias(self.TECTON_INTERNAL_SECONDARY_KEY_SET_COL)
        ]
        df = join_spec.join_and_aggregate(aggregations)
        res = df.withColumn(
            self.secondary_key, F.explode_outer(F.col(self.TECTON_INTERNAL_SECONDARY_KEY_SET_COL))
        ).drop(self.TECTON_INTERNAL_SECONDARY_KEY_SET_COL)

        return res


@attrs.frozen
class AggregationSecondaryKeyRollupSparkNode(SparkExecNode):
    """
    Rollup aggregation secondary key and corresponding feature values for each distinct window.

    The input node is a full aggregation node, which includes join keys and a secondary key. This node will:
        1. Group aggregation features by windows and put them into a temporary struct column for each distinct window.
        2. Rollup the temporal struct into a list by grouping on join keys and anchor time.
        3. Filter out secondary keys that do not appear in a particular window, and sort the list by secondary key.
        4. Transform each field in the temporal struct columns into its own column.

    E.g. Let's say a FV has a join_key `user_id` and a secondary key `ad_id` and the full aggregation node is like:
        | user_id | ad_id | value_max_1d     | value_max_3d     |  anchor_time | indicator_1d | indicator_3d |
        |---------|-------|------------------|------------------|--------------|--------------|--------------|
        | a       | 1     | 1                | 11               | 1            | 1            | 1            |
        | a       | 2     | 2                | 12               | 1            | 1            | 1            |
        | a       | 3     | None             | 13               | 1            | None         | 1            |
        | a       | 1     | 1                | 21               | 2            | 1            | 1            |
        | a       | 2     | 2                | 22               | 2            | 1            | 1            |

        The output of this node will be:
        | user_id | ad_id_1d    | ad_id_3d    | value_max_1d | value_max_3d | anchor_time |
        |---------|-------------|-------------|--------------|--------------|-------------|
        | a       | [1, 2]      | [1, 2, 3]   | [1, 2]       | [11, 12, 13] | 1           |
        | a       | [1, 2]      | [1, 2]      | [1, 2]       | [21, 22]     | 2           |

    Attributes:
        full_aggregation_node: The input node, which is always a full aggregation node.
        fdw: The feature definition wrapper that contains the useful metadata.
        group_by_columns: The columns to group by when rolling up the secondary key and its corresponding feature values.
    """

    full_aggregation_node: SparkExecNode
    fdw: FeatureDefinitionWrapper
    group_by_columns: Tuple[str, ...]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = self.full_aggregation_node.to_dataframe(spark)

        window_to_features = self.fdw.window_to_features_map_for_secondary_keys
        # Put all columns in each distinct window into a struct column with name "_struct_{window}", and then collect
        # all struct values in each column into a list by grouping on join keys and anchor time.
        temp_struct_columns = [
            F.struct(*cols).alias(temp_struct_column_name(window)) for window, cols in window_to_features.items()
        ]
        temp_struct_column_names = [temp_struct_column_name(window) for window in window_to_features.keys()]

        # Rollup aggregation secondary keys and corresponding feature values by collecting struct values into a list so
        # that for each unique join keys and anchor time pair, we have a list of seoncdary keys and features for each
        # distinct window.
        df = (
            df.select(*df.columns, *temp_struct_columns)
            .groupBy(*self.group_by_columns)
            .agg(*[F.collect_list(c).alias(c) for c in temp_struct_column_names])
        )
        return self._transform_temp_struct_to_feature_columns(df)

    def _transform_temp_struct_to_feature_columns(self, df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
        for secondary_key_output in self.fdw.materialized_fv_spec.secondary_key_rollup_outputs:
            time_window = secondary_key_output.time_window
            temp_struct_column = temp_struct_column_name(time_window)

            # Filter out secondary keys that don't appear in a given window
            df = self._filter_aggregation_secondary_key_by_indicator(df, time_window, temp_struct_column)

            # Sort the secondary key values in ascending order.
            sort_function = f"(left, right) -> case when left.{self.fdw.aggregation_secondary_key} < right.{self.fdw.aggregation_secondary_key} then -1 when left.{self.fdw.aggregation_secondary_key} > right.{self.fdw.aggregation_secondary_key} then 1 else 0 end"
            df = df.withColumn(temp_struct_column, F.expr(f"array_sort({temp_struct_column}, {sort_function})"))

            # Extract each feature value in the struct into its own column.
            df = self._extract_features_columns_from_struct_columns(df, time_window, temp_struct_column)

            # Extract the secondary key values in the struct into its own column.
            df = df.withColumn(
                secondary_key_output.name,
                F.expr(f"transform({temp_struct_column}, x -> x.{self.fdw.aggregation_secondary_key})"),
            )

            df = df.drop(temp_struct_column)
        return df

    def _filter_aggregation_secondary_key_by_indicator(
        self, df: pyspark.sql.DataFrame, window: TimeWindowSpec, temp_struct_column_name: str
    ) -> pyspark.sql.DataFrame:
        # Filter out those secondary keys with their corresponding feature values in this time window if the
        # indicator column is `null`. We use a window based `sum` on the aggregation secondary key in
        # `AsofJoinFullAggNode` so that the indicator column is `null` for those secondary keys that are not present.
        return df.withColumn(
            temp_struct_column_name,
            F.filter(
                temp_struct_column_name,
                lambda x: x.getField(temp_indictor_column_name(window)).isNotNull(),
            ),
        )

    def _extract_features_columns_from_struct_columns(
        self, df: pyspark.sql.DataFrame, window: TimeWindowSpec, temp_struct_column_name: str
    ) -> pyspark.sql.DataFrame:
        for agg_feature in self.fdw.trailing_time_window_aggregation().features:
            if create_time_window_spec_from_data_proto(agg_feature.time_window) == window:
                df = df.withColumn(
                    agg_feature.output_feature_name,
                    F.expr(f"transform({temp_struct_column_name}, x -> x.{agg_feature.output_feature_name})"),
                )
        return df


@attrs.frozen
class InnerJoinOnRangeSparkNode(SparkExecNode):
    """Joins the left against the right, using the right columns as the conditional.

    In pseudo-sql it is:
        SELECT
            left.*
            right.*
        FROM
            left INNER JOIN right
            ON (right.right_inclusive_start_column IS null OR left.left_join_condition_column >= right.right_inclusive_start_column)
                AND left.left_join_condition_column < right.right_exclusive_end_column)
    """

    left: SparkExecNode
    right: SparkExecNode
    left_join_condition_column: str
    right_inclusive_start_column: str
    right_exclusive_end_column: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        left_df = self.left.to_dataframe(spark).alias("left_join_df")
        right_df = self.right.to_dataframe(spark).alias("right_join_df")

        join_condition = (
            # If start_time is not null, use the regular condition
            (
                F.isnull(right_df[self.right_inclusive_start_column])
                | (left_df[self.left_join_condition_column] >= right_df[self.right_inclusive_start_column])
            )
            & (left_df[self.left_join_condition_column] < right_df[self.right_exclusive_end_column])
        )

        # NOTE: the broadcast is important since the Spark query optimizer is not able to infer it on DataFrames defined in memory (as of Spark 3.2).
        return left_df.join(F.broadcast(right_df), join_condition, "inner")


@attrs.frozen
class TakeLastRowSparkNode(SparkExecNode):
    input_node: SparkExecNode
    partition_by_columns: Tuple[str, ...]
    order_by_column: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark).alias("input_node_df")
        window = spark_window.Window.partitionBy(*self.partition_by_columns).orderBy(F.col(self.order_by_column).desc())
        row_number_col = "__tecton_row_num"
        assert row_number_col not in input_df.columns
        df = input_df.withColumn(row_number_col, F.row_number().over(window))
        df = df.filter(F.col(row_number_col) == 1).drop(row_number_col)
        return df


@attrs.frozen
class TemporalBatchTableFormatSparkNode(SparkExecNode):
    input_node: SparkExecNode
    fdw: FeatureDefinitionWrapper = attrs.field()
    online_batch_table_part: OnlineBatchTablePart

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark).alias("input_node_df")
        df = input_df.select(
            *self.fdw.join_keys,
            F.lit(self.online_batch_table_part.window_index)
            .cast("long")
            .alias(aggregation_group_id()),  # Cast for equivalence with aggregate QT.
            F.array(F.struct(*self.online_batch_table_part.schema.column_names())).alias(
                str(self.online_batch_table_part.window_index)
            ),
        )
        return df


@attrs.frozen
class ExplodeTimestampByTimeWindowsSparkNode(SparkExecNode):
    """
    Explodes each timestamp in 'column_name' into multiple rows, each with a new timestamp
    that is the sum of the time and the time window.
    """

    input_node: SparkExecNode
    timestamp_field: str
    fdw: FeatureDefinitionWrapper
    time_filter: pendulum.Period

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        spine_df = self.input_node.to_dataframe(spark)
        feature_start_time = self.fdw.feature_start_timestamp

        time_deltas = set()
        for time_window in self._time_window_specs():
            if isinstance(time_window, LifetimeWindowSpec):
                time_deltas.add(timedelta(0))
            elif isinstance(time_window, TimeWindowSeriesSpec):
                for time_window in time_window.time_windows:
                    time_deltas.update((time_window.window_start, time_window.window_end))
            elif isinstance(time_window, RelativeTimeWindowSpec):
                time_deltas.update((time_window.window_start, time_window.window_end))
            else:
                msg = f"Invalid time_window type: {type(time_window)}"
                raise ValueError(msg)

        time_deltas_ns = [
            convert_timedelta_for_version(abs(td), self.fdw.get_feature_store_format_version) for td in time_deltas
        ]

        new_anchor_times = F.array(*(F.col(anchor_time()) + F.lit(td) for td in time_deltas_ns))
        df_exploded = spine_df.withColumn(anchor_time(), F.explode(new_anchor_times))

        start_time_ns = convert_timestamp_for_version(self.time_filter.start, self.fdw.get_feature_store_format_version)
        df_exploded = df_exploded.withColumn(
            anchor_time(),
            F.when(
                (F.col(anchor_time()) <= start_time_ns),
                F.lit(start_time_ns),
            ).otherwise(F.col(anchor_time())),
        )

        end_time_ns = convert_timestamp_for_version(self.time_filter.end, self.fdw.get_feature_store_format_version)
        df_exploded = df_exploded.filter(df_exploded[anchor_time()] < end_time_ns)

        return df_exploded.dropDuplicates()

    def _time_window_specs(self) -> List[TimeWindowSpec]:
        return [
            create_time_window_spec_from_data_proto(feature.time_window)
            for feature in self.fdw.materialized_fv_spec.aggregate_features
        ]


@attrs.frozen
class AsofJoinSawtoothAggSparkNode(SparkExecNode):
    batch_input_node: SparkExecNode
    stream_input_node: SparkExecNode
    spine_input_node: SparkExecNode
    sawtooth_aggregation_data: compaction_utils.SawtoothAggregationData
    fdw: FeatureDefinitionWrapper

    def _add_output_feature_name_cols(self, stream_df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
        # Match the schema of the batch input node which has a column for each output feature name.
        time_aggregation = self.fdw.trailing_time_window_aggregation()
        new_columns = []
        cols_to_drop = []
        for feature in time_aggregation.features:
            col_datatype = self.fdw.view_schema.to_dict()[feature.input_feature_name]
            aggregation_plan = get_aggregation_plan(
                feature.function,
                feature.function_params,
                time_aggregation.is_continuous,
                time_aggregation.time_key,
                col_datatype,
            )
            feature_names = aggregation_plan.materialized_column_names(feature.input_feature_name)
            new_columns.extend([F.col(name).alias(feature.output_feature_name) for name in feature_names])
            cols_to_drop += feature_names

        columns_to_keep = [col for col in stream_df.columns if col not in cols_to_drop]
        return stream_df.select(*columns_to_keep, *new_columns)

    def _union_batch_and_stream(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        left_df = self.batch_input_node.to_dataframe(spark)
        right_df = self.stream_input_node.to_dataframe(spark)
        right_df = self._add_output_feature_name_cols(right_df)

        common_cols = set(left_df.columns) & set(right_df.columns)
        left_only_columns = set(left_df.columns) - common_cols
        right_only_columns = set(right_df.columns) - common_cols

        left_full_cols = (
            [F.col(x) for x in common_cols]
            + [F.col(x) for x in left_only_columns]
            + [F.lit(None).alias(x) for x in right_only_columns]
        )
        right_full_cols = (
            [F.col(x) for x in common_cols]
            + [F.lit(None).alias(x) for x in left_only_columns]
            + [F.col(x) for x in right_only_columns]
        )
        left_df = left_df.select(left_full_cols)
        right_df = right_df.select(right_full_cols)
        return left_df.union(right_df)

    def _add_time_partition_columns(self, input_df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
        output_df = input_df
        for sawtooth_size in self.sawtooth_aggregation_data.get_unique_sawtooth_sizes():
            anchor_time_column = self.sawtooth_aggregation_data.get_anchor_time_column_for_sawtooth_size(sawtooth_size)
            partition_col = self.sawtooth_aggregation_data.get_time_partition_column_for_sawtooth_size(sawtooth_size)
            interval_seconds = time_utils.convert_timedelta_for_version(
                sawtooth_size, self.fdw.get_feature_store_format_version
            )
            output_df = output_df.withColumn(
                partition_col, F.col(anchor_time_column) - F.col(anchor_time_column) % interval_seconds
            )
        return output_df

    def _drop_time_partition_columns(self, input_df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
        output_df = input_df
        for sawtooth_size in self.sawtooth_aggregation_data.get_unique_sawtooth_sizes():
            partition_col = self.sawtooth_aggregation_data.get_time_partition_column_for_sawtooth_size(sawtooth_size)
            output_df = output_df.drop(partition_col)
        return output_df

    def _get_aggregations(self, join_spec: _AsofJoinerAndWindowAggregator) -> List[pyspark.sql.Column]:
        time_aggregation = self.fdw.trailing_time_window_aggregation()
        aggregations = []
        for feature in time_aggregation.features:
            col_datatype = self.fdw.view_schema.to_dict()[feature.input_feature_name]
            aggregation_plan = get_aggregation_plan(
                feature.function,
                feature.function_params,
                time_aggregation.is_continuous,
                time_aggregation.time_key,
                col_datatype,
            )
            time_window = create_time_window_spec_from_data_proto(feature.time_window)
            if isinstance(time_window, TimeWindowSeriesSpec):
                msg = "TimeWindowSeriesSpec is not supported for sawtooth aggregation"
                raise ValueError(msg)

            sawtooth_tile_size = time_utils.proto_to_duration(feature.batch_sawtooth_tile_size)
            timestamp_col = self.sawtooth_aggregation_data.get_anchor_time_column_for_sawtooth_size(sawtooth_tile_size)
            time_partition_col = self.sawtooth_aggregation_data.get_time_partition_column_for_sawtooth_size(
                sawtooth_tile_size
            )
            identifier_partition_col = self.sawtooth_aggregation_data.get_identifier_partition_column_for_sawtooth_size(
                sawtooth_tile_size
            )
            window_spec = join_spec.get_range_between_window_spec(
                spark_window.Window.unboundedPreceding,
                spark_window.Window.currentRow,
                timestamp_col=timestamp_col,
                extra_partition_cols=[time_partition_col, identifier_partition_col],
            )
            agg = aggregation_plan.full_aggregation_transform([feature.output_feature_name], window_spec)
            aggregations.append(agg.alias(feature.output_feature_name))
        return aggregations

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        spine_df = self._add_time_partition_columns(self.spine_input_node.to_dataframe(spark))
        events_df = self._union_batch_and_stream(spark)
        events_df = self._add_time_partition_columns(events_df)

        timestamp_cols = self.sawtooth_aggregation_data.get_anchor_time_columns()
        partition_columns = self.fdw.join_keys
        agg_specific_partition_columns = (
            self.sawtooth_aggregation_data.get_identifier_partition_columns()
            + self.sawtooth_aggregation_data.get_truncated_anchor_time_partition_columns()
        )

        join_spec = _AsofJoinerAndWindowAggregator.create(
            left_df=spine_df,
            right_df=events_df,
            left_ts_cols=timestamp_cols,
            right_ts_cols=timestamp_cols,
            partition_cols=partition_columns,
            extra_partition_cols=agg_specific_partition_columns,
            use_window_range_between_value=False,
        )
        aggregations = self._get_aggregations(join_spec)
        result_df = join_spec.join_and_aggregate(aggregations)
        return self._drop_time_partition_columns(result_df)



# This is Spark-free module, which is used by both Spark and Ray materializations
import abc
import datetime
import logging
import threading
from typing import Callable
from typing import Optional
from typing import Tuple

from tecton_proto.materialization.job_metadata__client_pb2 import JobMetadata
from tecton_proto.materialization.job_metadata__client_pb2 import JobMetadataTableType
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams


logger = logging.getLogger(__name__)
# This section of constants should be used purely for ensuring idempotence of spark jobs.
IDEMPOTENCE_KEY_ATTRIBUTE = "idempotence_key"
VALUE_ATTRIBUTE = "value"
TTL_ATTRIBUTE = "ttl"
LAST_UPDATED_ATTRIBUTE = "last_updated"
RUN_ID_PREFIX = "id:"

TTL_DURATION_SECONDS = int(datetime.timedelta(days=60).total_seconds())

JOB_EXEC_PKEY_ATTRIBUTE = "id"
JOB_EXEC_LAST_UPDATED_ATTRIBUTE = "last_updated"
JOB_EXEC_DATA_ATTRIBUTE = "data"
JOB_EXEC_VERSION_ATTRIBUTE = "version"
CONSUMPTION_BUCKET_SIZE = datetime.timedelta(hours=1)  # see ConsumptionConstants.kt


class JobMetadataClient(abc.ABC):
    @staticmethod
    def for_params(materialization_task_params: MaterializationTaskParams) -> "JobMetadataClient":
        if materialization_task_params.job_metadata_table_type == JobMetadataTableType.JOB_METADATA_TABLE_TYPE_GCS:
            from tecton_materialization.common.job_metadata_gcp import GCSMetadataClient

            return GCSMetadataClient.for_params(materialization_task_params)
        elif materialization_task_params.job_metadata_table_type == JobMetadataTableType.JOB_METADATA_TABLE_TYPE_DYNAMO:
            from tecton_materialization.common.job_metadata_aws import DynamoMetadataClient

            return DynamoMetadataClient.for_params(materialization_task_params)
        else:
            msg = f"Unhandled JobMetadataTableType: {materialization_task_params.job_metadata_table_type}"
            raise Exception(msg)

    @abc.abstractmethod
    def get(self) -> Tuple[JobMetadata, int]:
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, updater: Callable[[JobMetadata], Optional[JobMetadata]]) -> JobMetadata:
        raise NotImplementedError


class LazyJobMetadataClient(JobMetadataClient):
    def __init__(self, params: MaterializationTaskParams):
        self._params = params
        self._delegate = None
        self._lock = threading.Lock()

    def _get_delegate(self):
        with self._lock:
            if self._delegate is None:
                self._delegate = JobMetadataClient.for_params(self._params)
            return self._delegate

    def get(self) -> Tuple[JobMetadata, int]:
        return self._get_delegate().get()

    def update(self, updater: Callable[[JobMetadata], Optional[JobMetadata]]) -> JobMetadata:
        return self._get_delegate().update(updater)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/spark_common/libraries__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.spark_common import clusters__client_pb2 as tecton__proto_dot_spark__common_dot_clusters____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1tecton_proto/spark_common/libraries__client.proto\x12\x19tecton_proto.spark_common\x1a\x30tecton_proto/spark_common/clusters__client.proto\"\xca\x03\n\x07Library\x12\x12\n\x03jar\x18\x01 \x01(\tH\x00R\x03jar\x12\x12\n\x03\x65gg\x18\x02 \x01(\tH\x00R\x03\x65gg\x12\x12\n\x03whl\x18\x03 \x01(\tH\x00R\x03whl\x12P\n\x0cjar_resource\x18\x06 \x01(\x0b\x32+.tecton_proto.spark_common.ResourceLocationH\x00R\x0bjarResource\x12P\n\x0c\x65gg_resource\x18\x07 \x01(\x0b\x32+.tecton_proto.spark_common.ResourceLocationH\x00R\x0b\x65ggResource\x12P\n\x0cwhl_resource\x18\x08 \x01(\x0b\x32+.tecton_proto.spark_common.ResourceLocationH\x00R\x0bwhlResource\x12?\n\x05maven\x18\x04 \x01(\x0b\x32\'.tecton_proto.spark_common.MavenLibraryH\x00R\x05maven\x12<\n\x04pypi\x18\x05 \x01(\x0b\x32&.tecton_proto.spark_common.PyPiLibraryH\x00R\x04pypiB\x0e\n\x0clibrary_type\";\n\x0bPyPiLibrary\x12\x18\n\x07package\x18\x01 \x01(\tR\x07package\x12\x12\n\x04repo\x18\x02 \x01(\tR\x04repo\"d\n\x0cMavenLibrary\x12 \n\x0b\x63oordinates\x18\x01 \x01(\tR\x0b\x63oordinates\x12\x12\n\x04repo\x18\x02 \x01(\tR\x04repo\x12\x1e\n\nexclusions\x18\x03 \x03(\tR\nexclusionsB\x1b\n\x17\x63om.tecton.spark_commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.spark_common.libraries__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\027com.tecton.spark_commonP\001'
  _LIBRARY._serialized_start=131
  _LIBRARY._serialized_end=589
  _PYPILIBRARY._serialized_start=591
  _PYPILIBRARY._serialized_end=650
  _MAVENLIBRARY._serialized_start=652
  _MAVENLIBRARY._serialized_end=752
# @@protoc_insertion_point(module_scope)



import logging

from tecton_core import errors


logger = logging.getLogger(__name__)


class SnowflakeContext:
    """
    Get access to Snowflake connection and session.
    """

    _current_context_instance = None
    _session = None
    _connection = None

    def __init__(self, connection):
        self._connection = connection
        from snowflake.snowpark import Session

        connection_parameters = {
            "connection": connection,
        }
        self._session = Session.builder.configs(connection_parameters).create()

    def get_session(self):
        if self._session is None:
            raise errors.SNOWFLAKE_CONNECTION_NOT_SET
        return self._session

    def get_connection(self):
        if self._connection is None:
            raise errors.SNOWFLAKE_CONNECTION_NOT_SET
        return self._connection

    @classmethod
    def is_initialized(cls):
        return cls._current_context_instance is not None

    @classmethod
    def get_instance(cls) -> "SnowflakeContext":
        """
        Get the singleton instance of SnowflakeContext.
        """
        # If the instance doesn't exist, raise the error to instruct user to set connection first. Otherwise
        # return the current snowflake context.
        if cls._current_context_instance is not None:
            return cls._current_context_instance
        else:
            raise errors.SNOWFLAKE_CONNECTION_NOT_SET

    @classmethod
    def set_connection(cls, connection) -> "SnowflakeContext":  # noqa: ANN001
        logger.debug("Generating new Snowflake session")
        # validate snowflake connection
        if not connection.database:
            msg = "database"
            raise errors.MISSING_SNOWFAKE_CONNECTION_REQUIREMENTS(msg)
        if not connection.warehouse:
            msg = "warehouse"
            raise errors.MISSING_SNOWFAKE_CONNECTION_REQUIREMENTS(msg)
        if not connection.schema:
            msg = "schema"
            raise errors.MISSING_SNOWFAKE_CONNECTION_REQUIREMENTS(msg)

        cls._current_context_instance = cls(connection)
        return cls._current_context_instance



from unittest import TestCase
from unittest import mock

from google.protobuf.empty_pb2 import Empty

from tecton._internals.metadata_service_impl.auth_lib import InternalAuthProvider
from tecton._internals.metadata_service_impl.request_lib import InternalRequestProvider
from tecton._internals.metadata_service_impl.service_modules import GRPC_SERVICE_MODULES
from tecton_core.errors import FailedPreconditionError
from tecton_core.errors import TectonAbortedError
from tecton_core.errors import TectonAlreadyExistsError
from tecton_core.errors import TectonAPIInaccessibleError
from tecton_core.errors import TectonAPIValidationError
from tecton_core.errors import TectonDeadlineExceededError
from tecton_core.errors import TectonNotFoundError
from tecton_core.errors import TectonNotImplementedError
from tecton_core.errors import TectonOperationCancelledError
from tecton_core.errors import TectonResourceExhaustedError
from tecton_core.metadata_service_impl import error_lib
from tecton_core.metadata_service_impl import http_client
from tecton_core.metadata_service_impl.service_calls import GrpcCall


@mock.patch(
    "tecton._internals.metadata_service_impl.request_lib.InternalRequestProvider.request_url",
    return_value="https://test.tecton.ai/api",
)
@mock.patch("tecton.identities.okta.get_token_refresh_if_needed", lambda: None)
class HttpClientTest(TestCase):
    @mock.patch("tecton_core.metadata_service_impl.http_client._InternalHTTPStub.execute")
    def test_valid_request(self, mock_execute, _):
        """
        Tests the translation of the PureHTTPStub Nop(proto) method to _InternalHTTPStub.execute('Nop', proto).
        """
        expected_grpc_call = GrpcCall(
            "/tecton_proto.metadataservice.MetadataService/Nop", Empty.SerializeToString, Empty.FromString
        )
        mock_execute.return_value = Empty()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        response = stub.Nop(Empty())
        mock_execute.assert_called_once()
        mock_execute.assert_called_with(expected_grpc_call, Empty(), 300.0)
        assert response == Empty()

    @mock.patch("tecton_core.metadata_service_impl.http_client._InternalHTTPStub.execute")
    def test_timeout_param(self, mock_execute, _):
        """
        Test passing in a timeout to a method called on PureHTTPStub
        """
        expected_grpc_call = GrpcCall(
            "/tecton_proto.metadataservice.MetadataService/Nop", Empty.SerializeToString, Empty.FromString
        )
        mock_execute.return_value = Empty()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        response = stub.Nop(Empty(), timeout_sec=5.0)
        mock_execute.assert_called_once()
        mock_execute.assert_called_with(expected_grpc_call, Empty(), 5.0)
        assert response == Empty()

    @mock.patch("tecton_core.metadata_service_impl.http_client._InternalHTTPStub.execute")
    def test_invalid_method(self, mock_execute, _):
        """
        Tests error handling of a method called on PureHTTPStub that doesn't map to a valid MetadataService method.
        """
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaisesRegex(AttributeError, "Nonexistent MetadataService method: InvalidRequestName"):
            stub.InvalidRequestName(Empty())
        mock_execute.assert_not_called()

    @mock.patch("requests.Session.request")
    def test_unauthenticated_request(self, mock_request, _):
        """
        Test when unauthenticated status is returned
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.UNAUTHENTICATED.value, "detail": "thedetail"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(PermissionError) as context:
            stub.Nop(Empty())
        assert "Tecton credentials are invalid, not configured, or expired" in str(context.exception)

    @mock.patch("requests.Session.request")
    def test_permission_denied_request(self, mock_request, _):
        """
        Test when permission denied (unauthorized) status is returned
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.PERMISSION_DENIED.value, "detail": "thedetail"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaisesRegex(PermissionError, ".*Insufficient permissions.*"):
            stub.Nop(Empty())

    @mock.patch("requests.Session.request")
    def test_permission_denied_unauthenticated_request(self, mock_request, _):
        """
        Test when permission denied (unauthorized) status is returned
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {
                    "status": {
                        "code": error_lib.gRPCStatus.PERMISSION_DENIED.value,
                        "detail": "UNAUTHENTICATED: InvalidToken",
                    }
                }

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaisesRegex(PermissionError, ".*Insufficient permissions.*"):
            stub.Nop(Empty())

    # TODO: Remove with https://tecton.atlassian.net/browse/TEC-9107
    #  (once the metadata service no longer returns PERMISSION_DENIED when authentication is required but not included)
    @mock.patch("requests.Session.request")
    def test_permission_denied_no_header_request(self, mock_request, _):
        """
        Test when permission denied (unauthorized) status is returned
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.PERMISSION_DENIED.value, "detail": "thedetail"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaisesRegex(PermissionError, ".*Insufficient permissions.*"):
            stub.Nop(Empty())

    @mock.patch("requests.Session.request")
    def test_cancelled_request(self, mock_request, _):
        """
        Test when a request is cancelled
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.CANCELLED.value, "detail": "Operation cancelled"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonOperationCancelledError) as context:
            stub.Nop(Empty())
        self.assertIn("Operation cancelled", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_deadline_exceeded_request(self, mock_request, _):
        """
        Test when a request deadline is exceeded
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.DEADLINE_EXCEEDED.value, "detail": "Deadline exceeded"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonDeadlineExceededError) as context:
            stub.Nop(Empty())
        self.assertIn("Deadline exceeded", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_already_exists_request(self, mock_request, _):
        """
        Test when a resource already exists
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {
                    "status": {"code": error_lib.gRPCStatus.ALREADY_EXISTS.value, "detail": "Resource already exists"}
                }

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonAlreadyExistsError) as context:
            stub.Nop(Empty())
        self.assertIn("Resource already exists", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_resource_exhausted_request(self, mock_request, _):
        """
        Test when a resource is exhausted
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {
                    "status": {"code": error_lib.gRPCStatus.RESOURCE_EXHAUSTED.value, "detail": "Resource exhausted"}
                }

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonResourceExhaustedError) as context:
            stub.Nop(Empty())
        self.assertIn("Resource exhausted", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_aborted_request(self, mock_request, _):
        """
        Test when a request is aborted
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.ABORTED.value, "detail": "Operation aborted"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonAbortedError) as context:
            stub.Nop(Empty())
        self.assertIn("Operation aborted", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_out_of_range_request(self, mock_request, _):
        """
        Test when a value is out of range
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.OUT_OF_RANGE.value, "detail": "Value out of range"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonAPIValidationError) as context:
            stub.Nop(Empty())
        self.assertIn("Value out of range", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_unavailable_request(self, mock_request, _):
        """
        Test when the API is unavailable
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.UNAVAILABLE.value, "detail": "API unavailable"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonAPIInaccessibleError) as context:
            stub.Nop(Empty())
        self.assertIn("API unavailable", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_invalid_argument_request(self, mock_request, _):
        """
        Test when an invalid argument is provided
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.INVALID_ARGUMENT.value, "detail": "Invalid argument"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonAPIValidationError) as context:
            stub.Nop(Empty())
        self.assertIn("Invalid argument", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_failed_precondition_request(self, mock_request, _):
        """
        Test when a precondition fails
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {
                    "status": {"code": error_lib.gRPCStatus.FAILED_PRECONDITION.value, "detail": "Precondition failed"}
                }

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(FailedPreconditionError) as context:
            stub.Nop(Empty())
        self.assertIn("Precondition failed", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_not_found_request(self, mock_request, _):
        """
        Test when a resource is not found
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {"status": {"code": error_lib.gRPCStatus.NOT_FOUND.value, "detail": "Resource not found"}}

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonNotFoundError) as context:
            stub.Nop(Empty())
        self.assertIn("Resource not found", str(context.exception))

    @mock.patch("requests.Session.request")
    def test_unimplemented_request(self, mock_request, _):
        """
        Test when a feature is not implemented
        """

        class FakeResponse:
            def raise_for_status(self):
                pass

            def json(self):
                return {
                    "status": {"code": error_lib.gRPCStatus.UNIMPLEMENTED.value, "detail": "Feature not implemented"}
                }

        mock_request.return_value = FakeResponse()
        stub = http_client.PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
        with self.assertRaises(TectonNotImplementedError) as context:
            stub.Nop(Empty())
        self.assertIn("Feature not implemented", str(context.exception))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/framework_version__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n3tecton_proto/common/framework_version__client.proto\x12\x13tecton_proto.common*I\n\x10\x46rameworkVersion\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04\x46WV3\x10\x03\x12\x08\n\x04\x46WV5\x10\x05\"\x04\x08\x01\x10\x01\"\x04\x08\x02\x10\x02\"\x04\x08\x04\x10\x04\x42\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.framework_version__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _FRAMEWORKVERSION._serialized_start=76
  _FRAMEWORKVERSION._serialized_end=149
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/id__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.auditlog import metadata__client_pb2 as tecton__proto_dot_auditlog_dot_metadata____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$tecton_proto/common/id__client.proto\x12\x13tecton_proto.common\x1a,tecton_proto/auditlog/metadata__client.proto\"\x80\x01\n\x02Id\x12;\n\x15most_significant_bits\x18\x01 \x01(\x06\x42\x07\xea\x99\xbd\x46\x02\x08\x02R\x13mostSignificantBits\x12=\n\x16least_significant_bits\x18\x02 \x01(\x06\x42\x07\xea\x99\xbd\x46\x02\x08\x02R\x14leastSignificantBitsB\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.id__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _ID.fields_by_name['most_significant_bits']._options = None
  _ID.fields_by_name['most_significant_bits']._serialized_options = b'\352\231\275F\002\010\002'
  _ID.fields_by_name['least_significant_bits']._options = None
  _ID.fields_by_name['least_significant_bits']._serialized_options = b'\352\231\275F\002\010\002'
  _ID._serialized_start=108
  _ID._serialized_end=236
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pyspark import SparkContext
from pyspark.sql.column import Column, _to_java_column


def vector_to_array(col, dtype="float64"):
    """
    Converts a column of MLlib sparse/dense vectors into a column of dense arrays.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    col : :py:class:`pyspark.sql.Column` or str
        Input column
    dtype : str, optional
        The data type of the output array. Valid values: "float64" or "float32".

    Returns
    -------
    :py:class:`pyspark.sql.Column`
        The converted column of dense arrays.

    Examples
    --------
    >>> from pyspark.ml.linalg import Vectors
    >>> from pyspark.ml.functions import vector_to_array
    >>> from pyspark.mllib.linalg import Vectors as OldVectors
    >>> df = spark.createDataFrame([
    ...     (Vectors.dense(1.0, 2.0, 3.0), OldVectors.dense(10.0, 20.0, 30.0)),
    ...     (Vectors.sparse(3, [(0, 2.0), (2, 3.0)]),
    ...      OldVectors.sparse(3, [(0, 20.0), (2, 30.0)]))],
    ...     ["vec", "oldVec"])
    >>> df1 = df.select(vector_to_array("vec").alias("vec"),
    ...                 vector_to_array("oldVec").alias("oldVec"))
    >>> df1.collect()
    [Row(vec=[1.0, 2.0, 3.0], oldVec=[10.0, 20.0, 30.0]),
     Row(vec=[2.0, 0.0, 3.0], oldVec=[20.0, 0.0, 30.0])]
    >>> df2 = df.select(vector_to_array("vec", "float32").alias("vec"),
    ...                 vector_to_array("oldVec", "float32").alias("oldVec"))
    >>> df2.collect()
    [Row(vec=[1.0, 2.0, 3.0], oldVec=[10.0, 20.0, 30.0]),
     Row(vec=[2.0, 0.0, 3.0], oldVec=[20.0, 0.0, 30.0])]
    >>> df1.schema.fields
    [StructField(vec,ArrayType(DoubleType,false),false),
    StructField(oldVec,ArrayType(DoubleType,false),false)]
    >>> df2.schema.fields
    [StructField(vec,ArrayType(FloatType,false),false),
    StructField(oldVec,ArrayType(FloatType,false),false)]
    """
    sc = SparkContext._active_spark_context
    return Column(
        sc._jvm.org.apache.spark.ml.functions.vector_to_array(_to_java_column(col), dtype))


def array_to_vector(col):
    """
    Converts a column of array of numeric type into a column of dense vectors in MLlib

    .. versionadded:: 3.1.0

    Parameters
    ----------
    col : :py:class:`pyspark.sql.Column` or str
        Input column

    Returns
    -------
    :py:class:`pyspark.sql.Column`
        The converted column of MLlib dense vectors.

    Examples
    --------
    >>> from pyspark.ml.functions import array_to_vector
    >>> df1 = spark.createDataFrame([([1.5, 2.5],),], schema='v1 array<double>')
    >>> df1.select(array_to_vector('v1').alias('vec1')).collect()
    [Row(vec1=DenseVector([1.5, 2.5]))]
    >>> df2 = spark.createDataFrame([([1.5, 3.5],),], schema='v1 array<float>')
    >>> df2.select(array_to_vector('v1').alias('vec1')).collect()
    [Row(vec1=DenseVector([1.5, 3.5]))]
    >>> df3 = spark.createDataFrame([([1, 3],),], schema='v1 array<int>')
    >>> df3.select(array_to_vector('v1').alias('vec1')).collect()
    [Row(vec1=DenseVector([1.0, 3.0]))]
    """
    sc = SparkContext._active_spark_context
    return Column(
        sc._jvm.org.apache.spark.ml.functions.array_to_vector(_to_java_column(col)))


def _test():
    import doctest
    from pyspark.sql import SparkSession
    import pyspark.ml.functions
    import sys
    globs = pyspark.ml.functions.__dict__.copy()
    spark = SparkSession.builder \
        .master("local[2]") \
        .appName("ml.functions tests") \
        .getOrCreate()
    sc = spark.sparkContext
    globs['sc'] = sc
    globs['spark'] = spark

    (failure_count, test_count) = doctest.testmod(
        pyspark.ml.functions, globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class FilterContext:
    """`FilterContext` is passed as an argument to the data source function when `supports_time_filtering` is set to `True`. Using these parameters enables optimized query patterns for improved performance:

     - The method `<data source>.get_dataframe()` can be invoked with the arguments `start_time` or `end_time`.
     - When defining a Feature View, a FilteredSource can be paired with a Data Source. The Feature View will then pass FilterContext into the Data Source Function

    Note that Data Source Functions are expected to implement their own filtering logic.

    Example function utilizing filter_context

    .. code-block:: python
        from tecton import spark_batch_config
        from pyspark.sql.functions import col

        @spark_batch_config(supports_time_filtering=True)
        def hive_data_source_function(spark, filter_context):
            spark.sql(f"USE {hive_db_name}")
            df = spark.table(user_hive_table)
            ts_column = "timestamp"

            # Data Source Function handles its own filtering logic here
            if filter_context:
                if filter_context.start_time:
                    df = df.where(col(ts_column) >= filter_context.start_time)
                if filter_context.end_time:
                    df = df.where(col(ts_column) < filter_context.end_time)
            return df
    """

    start_time: Optional[datetime]
    end_time: Optional[datetime]



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class ResourceInformation(object):

    """
    Class to hold information about a type of Resource. A resource could be a GPU, FPGA, etc.
    The array of addresses are resource specific and its up to the user to interpret the address.

    One example is GPUs, where the addresses would be the indices of the GPUs

    .. versionadded:: 3.0.0

    Parameters
    ----------
    name : str
        the name of the resource
    addresses : list
        an array of strings describing the addresses of the resource

    Notes
    -----
    This API is evolving.
    """

    def __init__(self, name, addresses):
        self._name = name
        self._addresses = addresses

    @property
    def name(self):
        return self._name

    @property
    def addresses(self):
        return self._addresses



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/feature_view__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.args import feature_view__client_pb2 as tecton__proto_dot_args_dot_feature__view____client__pb2
from tecton_proto.args import pipeline__client_pb2 as tecton__proto_dot_args_dot_pipeline____client__pb2
from tecton_proto.common import aggregation_function__client_pb2 as tecton__proto_dot_common_dot_aggregation__function____client__pb2
from tecton_proto.common import compute_mode__client_pb2 as tecton__proto_dot_common_dot_compute__mode____client__pb2
from tecton_proto.common import data_source_type__client_pb2 as tecton__proto_dot_common_dot_data__source__type____client__pb2
from tecton_proto.common import framework_version__client_pb2 as tecton__proto_dot_common_dot_framework__version____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2
from tecton_proto.common import time_window__client_pb2 as tecton__proto_dot_common_dot_time__window____client__pb2
from tecton_proto.data import fco_metadata__client_pb2 as tecton__proto_dot_data_dot_fco__metadata____client__pb2
from tecton_proto.data import fv_materialization__client_pb2 as tecton__proto_dot_data_dot_fv__materialization____client__pb2
from tecton_proto.data import realtime_compute__client_pb2 as tecton__proto_dot_data_dot_realtime__compute____client__pb2
from tecton_proto.modelartifactservice import model_artifact_service__client_pb2 as tecton__proto_dot_modelartifactservice_dot_model__artifact__service____client__pb2
from tecton_proto.validation import validator__client_pb2 as tecton__proto_dot_validation_dot_validator____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,tecton_proto/data/feature_view__client.proto\x12\x11tecton_proto.data\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a,tecton_proto/args/feature_view__client.proto\x1a(tecton_proto/args/pipeline__client.proto\x1a\x36tecton_proto/common/aggregation_function__client.proto\x1a.tecton_proto/common/compute_mode__client.proto\x1a\x32tecton_proto/common/data_source_type__client.proto\x1a\x33tecton_proto/common/framework_version__client.proto\x1a$tecton_proto/common/id__client.proto\x1a(tecton_proto/common/schema__client.proto\x1a-tecton_proto/common/time_window__client.proto\x1a,tecton_proto/data/fco_metadata__client.proto\x1a\x32tecton_proto/data/fv_materialization__client.proto\x1a\x30tecton_proto/data/realtime_compute__client.proto\x1a\x46tecton_proto/modelartifactservice/model_artifact_service__client.proto\x1a/tecton_proto/validation/validator__client.proto\"\xe2\x10\n\x0b\x46\x65\x61tureView\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12\x41\n\x0c\x66\x63o_metadata\x18\x02 \x01(\x0b\x32\x1e.tecton_proto.data.FcoMetadataR\x0b\x66\x63oMetadata\x12\x36\n\nentity_ids\x18\x03 \x03(\x0b\x32\x17.tecton_proto.common.IdR\tentityIds\x12\x1b\n\tjoin_keys\x18\x04 \x03(\tR\x08joinKeys\x12?\n\x07schemas\x18\x0b \x01(\x0b\x32%.tecton_proto.data.FeatureViewSchemasR\x07schemas\x12P\n\x0b\x65nrichments\x18\xe8\x07 \x01(\x0b\x32).tecton_proto.data.FeatureViewEnrichmentsB\x02\x18\x01R\x0b\x65nrichments\x12U\n\x12temporal_aggregate\x18\x07 \x01(\x0b\x32$.tecton_proto.data.TemporalAggregateH\x00R\x11temporalAggregate\x12\x39\n\x08temporal\x18\x08 \x01(\x0b\x32\x1b.tecton_proto.data.TemporalH\x00R\x08temporal\x12\\\n\x15realtime_feature_view\x18\x11 \x01(\x0b\x32&.tecton_proto.data.RealtimeFeatureViewH\x00R\x13realtimeFeatureView\x12\x46\n\rfeature_table\x18\x10 \x01(\x0b\x32\x1f.tecton_proto.data.FeatureTableH\x00R\x0c\x66\x65\x61tureTable\x12#\n\rtimestamp_key\x18\n \x01(\tR\x0ctimestampKey\x12W\n\x14online_serving_index\x18\x05 \x01(\x0b\x32%.tecton_proto.data.OnlineServingIndexR\x12onlineServingIndex\x12\x37\n\x08pipeline\x18\x06 \x01(\x0b\x32\x1b.tecton_proto.args.PipelineR\x08pipeline\x12\x62\n\x16materialization_params\x18\t \x01(\x0b\x32+.tecton_proto.data.NewMaterializationParamsR\x15materializationParams\x12\x37\n\x17materialization_enabled\x18\x0c \x01(\x08R\x16materializationEnabled\x12}\n!materialization_state_transitions\x18\r \x03(\x0b\x32\x31.tecton_proto.data.MaterializationStateTransitionR\x1fmaterializationStateTransitions\x12P\n\x11monitoring_params\x18\x0e \x01(\x0b\x32#.tecton_proto.data.MonitoringParamsR\x10monitoringParams\x12?\n\x1c\x66\x65\x61ture_store_format_version\x18\x17 \x01(\x05R\x19\x66\x65\x61tureStoreFormatVersion\x12G\n\x0esnowflake_data\x18\x18 \x01(\x0b\x32 .tecton_proto.data.SnowflakeDataR\rsnowflakeData\x12/\n\x11\x66ramework_version\x18\x13 \x01(\x05\x42\x02\x18\x01R\x10\x66rameworkVersion\x12H\n\nfw_version\x18\x1a \x01(\x0e\x32%.tecton_proto.common.FrameworkVersionB\x02\x18\x01R\tfwVersion\x12\x17\n\x07web_url\x18\x19 \x01(\tR\x06webUrl\x12H\n\rbatch_trigger\x18\x1b \x01(\x0e\x32#.tecton_proto.args.BatchTriggerTypeR\x0c\x62\x61tchTrigger\x12[\n\x0fvalidation_args\x18\x1c \x01(\x0b\x32\x32.tecton_proto.validation.FeatureViewValidationArgsR\x0evalidationArgs\x12T\n\x13\x64\x61ta_quality_config\x18\x1d \x01(\x0b\x32$.tecton_proto.data.DataQualityConfigR\x11\x64\x61taQualityConfig\x12\x45\n\x07options\x18\x1e \x03(\x0b\x32+.tecton_proto.data.FeatureView.OptionsEntryR\x07options\x12S\n\x12\x62\x61tch_compute_mode\x18\x1f \x01(\x0e\x32%.tecton_proto.common.BatchComputeModeR\x10\x62\x61tchComputeMode\x12L\n\x0c\x63\x61\x63he_config\x18  \x01(\x0b\x32).tecton_proto.data.FeatureViewCacheConfigR\x0b\x63\x61\x63heConfig\x12\x34\n\x16\x63ontext_parameter_name\x18! \x01(\tR\x14\x63ontextParameterName\x1a:\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x13\n\x11\x66\x65\x61ture_view_type\"\xe8\x03\n\x11TemporalAggregate\x12@\n\x0eslide_interval\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\rslideInterval\x12\x32\n\x15slide_interval_string\x18\x05 \x01(\tR\x13slideIntervalString\x12?\n\x08\x66\x65\x61tures\x18\x02 \x03(\x0b\x32#.tecton_proto.data.AggregateFeatureR\x08\x66\x65\x61tures\x12#\n\ris_continuous\x18\x03 \x01(\x08R\x0cisContinuous\x12M\n\x10\x64\x61ta_source_type\x18\x04 \x01(\x0e\x32#.tecton_proto.common.DataSourceTypeR\x0e\x64\x61taSourceType\x12:\n\x19\x61ggregation_secondary_key\x18\x06 \x01(\tR\x17\x61ggregationSecondaryKey\x12l\n\x1csecondary_key_output_columns\x18\x07 \x03(\x0b\x32+.tecton_proto.data.SecondaryKeyOutputColumnR\x19secondaryKeyOutputColumns\"\xd8\x03\n\x10\x41ggregateFeature\x12,\n\x12input_feature_name\x18\x01 \x01(\tR\x10inputFeatureName\x12.\n\x13output_feature_name\x18\x02 \x01(\tR\x11outputFeatureName\x12\x44\n\x08\x66unction\x18\x03 \x01(\x0e\x32(.tecton_proto.common.AggregationFunctionR\x08\x66unction\x12W\n\x0f\x66unction_params\x18\x05 \x01(\x0b\x32..tecton_proto.common.AggregationFunctionParamsR\x0e\x66unctionParams\x12\x31\n\x06window\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x06window\x12@\n\x0btime_window\x18\x06 \x01(\x0b\x32\x1f.tecton_proto.common.TimeWindowR\ntimeWindow\x12R\n\x18\x62\x61tch_sawtooth_tile_size\x18\x07 \x01(\x0b\x32\x19.google.protobuf.DurationR\x15\x62\x61tchSawtoothTileSize\"\x82\x01\n\x10\x45mbeddingFeature\x12*\n\x11input_column_name\x18\x01 \x01(\tR\x0finputColumnName\x12,\n\x12output_column_name\x18\x02 \x01(\tR\x10outputColumnName\x12\x14\n\x05model\x18\x05 \x01(\tR\x05model\"\xf3\x01\n\x10InferenceFeature\x12@\n\rinput_columns\x18\x01 \x03(\x0b\x32\x1b.tecton_proto.common.ColumnR\x0cinputColumns\x12@\n\routput_column\x18\x02 \x01(\x0b\x32\x1b.tecton_proto.common.ColumnR\x0coutputColumn\x12[\n\x0emodel_artifact\x18\x03 \x01(\x0b\x32\x34.tecton_proto.modelartifactservice.ModelArtifactInfoR\rmodelArtifact\"\xf5\x01\n\x1dTrailingTimeWindowAggregation\x12\x19\n\x08time_key\x18\x01 \x01(\tR\x07timeKey\x12S\n\x18\x61ggregation_slide_period\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x16\x61ggregationSlidePeriod\x12?\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32#.tecton_proto.data.AggregateFeatureR\x08\x66\x65\x61tures\x12#\n\ris_continuous\x18\x04 \x01(\x08R\x0cisContinuous\"\xc5\x03\n\x08Temporal\x12:\n\x0bserving_ttl\x18\x10 \x01(\x0b\x32\x19.google.protobuf.DurationR\nservingTtl\x12M\n\x10\x64\x61ta_source_type\x18\x11 \x01(\x0e\x32#.tecton_proto.common.DataSourceTypeR\x0e\x64\x61taSourceType\x12J\n\x0f\x62\x61\x63kfill_config\x18\x12 \x01(\x0b\x32!.tecton_proto.args.BackfillConfigR\x0e\x62\x61\x63kfillConfig\x12\x33\n\x15incremental_backfills\x18\x13 \x01(\x08R\x14incrementalBackfills\x12#\n\ris_continuous\x18\x14 \x01(\x08R\x0cisContinuous\x12\x43\n\nembeddings\x18\x15 \x03(\x0b\x32#.tecton_proto.data.EmbeddingFeatureR\nembeddings\x12\x43\n\ninferences\x18\x16 \x03(\x0b\x32#.tecton_proto.data.InferenceFeatureR\ninferences\"\x9a\x01\n\x0c\x46\x65\x61tureTable\x12%\n\x0eonline_enabled\x18\x01 \x01(\x08R\ronlineEnabled\x12\'\n\x0foffline_enabled\x18\x02 \x01(\x08R\x0eofflineEnabled\x12:\n\x0bserving_ttl\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\nservingTtl\"\x97\x01\n\x13RealtimeFeatureView\x12\x13\n\x05no_op\x18\x01 \x01(\x08R\x04noOp\x12\x65\n\x16supported_environments\x18\x03 \x03(\x0b\x32..tecton_proto.data.RemoteFunctionComputeConfigR\x15supportedEnvironmentsJ\x04\x08\x02\x10\x03\"\xc5\x02\n\x12\x46\x65\x61tureViewSchemas\x12<\n\x0bview_schema\x18\x01 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\nviewSchema\x12\x35\n\x17is_explicit_view_schema\x18\x04 \x01(\x08R\x14isExplicitViewSchema\x12R\n\x16materialization_schema\x18\x02 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x15materializationSchema\x12\x66\n\x19online_batch_table_format\x18\x03 \x01(\x0b\x32+.tecton_proto.common.OnlineBatchTableFormatR\x16onlineBatchTableFormat\"1\n\x12OnlineServingIndex\x12\x1b\n\tjoin_keys\x18\x01 \x03(\tR\x08joinKeys\"\xb3\x03\n\x1eMaterializationStateTransition\x12\x38\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\ttimestamp\x12%\n\x0eonline_enabled\x18\x02 \x01(\x08R\ronlineEnabled\x12\'\n\x0foffline_enabled\x18\x03 \x01(\x08R\x0eofflineEnabled\x12R\n\x17\x66\x65\x61ture_start_timestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x15\x66\x65\x61tureStartTimestamp\x12\x44\n\x1ematerialization_serial_version\x18\x05 \x01(\x05R\x1cmaterializationSerialVersion\x12\x37\n\x18\x66orce_stream_job_restart\x18\x06 \x01(\x08R\x15\x66orceStreamJobRestart\x12\x34\n\x16tecton_runtime_version\x18\x07 \x01(\tR\x14tectonRuntimeVersion\"\xaf\x01\n\x19ParquetOfflineStoreParams\x12I\n\x13time_partition_size\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11timePartitionSize\x12G\n\x07version\x18\x02 \x01(\x0e\x32-.tecton_proto.data.ParquetOfflineStoreVersionR\x07version\"\xab\x01\n\x17\x44\x65ltaOfflineStoreParams\x12I\n\x13time_partition_size\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11timePartitionSize\x12\x45\n\x07version\x18\x02 \x01(\x0e\x32+.tecton_proto.data.DeltaOfflineStoreVersionR\x07version\"\xac\x01\n\x12OfflineStoreParams\x12H\n\x07parquet\x18\x01 \x01(\x0b\x32,.tecton_proto.data.ParquetOfflineStoreParamsH\x00R\x07parquet\x12\x42\n\x05\x64\x65lta\x18\x02 \x01(\x0b\x32*.tecton_proto.data.DeltaOfflineStoreParamsH\x00R\x05\x64\x65ltaB\x08\n\x06params\"\xa0\x01\n FeaturePublishOfflineStoreConfig\x12\x32\n\x15publish_full_features\x18\x01 \x01(\x08R\x13publishFullFeatures\x12H\n\x12publish_start_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10publishStartTime\"\xf7\r\n\x18NewMaterializationParams\x12\x46\n\x11schedule_interval\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x10scheduleInterval\x12\x62\n\x1fmaterialization_start_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x1dmaterializationStartTimestamp\x12R\n\x17\x66\x65\x61ture_start_timestamp\x18\r \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x15\x66\x65\x61tureStartTimestamp\x12l\n%manual_trigger_backfill_end_timestamp\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.TimestampR!manualTriggerBackfillEndTimestamp\x12M\n\x15max_backfill_interval\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x13maxBackfillInterval\x12\x33\n\x16writes_to_online_store\x18\x04 \x01(\x08R\x13writesToOnlineStore\x12\x35\n\x17writes_to_offline_store\x18\x05 \x01(\x08R\x14writesToOfflineStore\x12^\n\x14offline_store_config\x18\x06 \x01(\x0b\x32,.tecton_proto.args.OfflineFeatureStoreConfigR\x12offlineStoreConfig\x12W\n\x14offline_store_params\x18\x0e \x01(\x0b\x32%.tecton_proto.data.OfflineStoreParamsR\x12offlineStoreParams\x12U\n\x15\x62\x61tch_materialization\x18\x07 \x01(\x0b\x32 .tecton_proto.args.ClusterConfigR\x14\x62\x61tchMaterialization\x12W\n\x16stream_materialization\x18\x08 \x01(\x0b\x32 .tecton_proto.args.ClusterConfigR\x15streamMaterialization\x12L\n\x15max_source_data_delay\x18\t \x01(\x0b\x32\x19.google.protobuf.DurationR\x12maxSourceDataDelay\x12T\n\x13online_store_params\x18\n \x01(\x0b\x32$.tecton_proto.data.OnlineStoreParamsR\x11onlineStoreParams\x12\x44\n\routput_stream\x18\x0b \x01(\x0b\x32\x1f.tecton_proto.args.OutputStreamR\x0coutputStream\x12]\n\x11time_range_policy\x18\x0c \x01(\x0e\x32\x31.tecton_proto.data.MaterializationTimeRangePolicyR\x0ftimeRangePolicy\x12\x64\n\x19online_backfill_load_type\x18\x10 \x01(\x0e\x32).tecton_proto.data.OnlineBackfillLoadTypeR\x16onlineBackfillLoadType\x12\x44\n\x1etecton_materialization_runtime\x18\x11 \x01(\tR\x1ctectonMaterializationRuntime\x12\x83\x01\n$feature_publish_offline_store_config\x18\x13 \x01(\x0b\x32\x33.tecton_proto.data.FeaturePublishOfflineStoreConfigR featurePublishOfflineStoreConfig\x12-\n\x12\x63ompaction_enabled\x18\x14 \x01(\x08R\x11\x63ompactionEnabled\x12\x32\n\x15stream_tiling_enabled\x18\x17 \x01(\x08R\x13streamTilingEnabled\x12 \n\x0b\x65nvironment\x18\x15 \x01(\tR\x0b\x65nvironment\x12\x43\n\x10stream_tile_size\x18\x16 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0estreamTileSizeJ\x04\x08\x12\x10\x13\"\xec\x01\n\x11OnlineStoreParams\x12@\n\x06\x64ynamo\x18\x02 \x01(\x0b\x32&.tecton_proto.data.DynamoDbOnlineStoreH\x00R\x06\x64ynamo\x12;\n\x05redis\x18\x03 \x01(\x0b\x32#.tecton_proto.data.RedisOnlineStoreH\x00R\x05redis\x12\x44\n\x08\x62igtable\x18\x04 \x01(\x0b\x32&.tecton_proto.data.BigtableOnlineStoreH\x00R\x08\x62igtableB\x0c\n\nstore_typeJ\x04\x08\x01\x10\x02\"\xa1\x02\n\x13\x44ynamoDbOnlineStore\x12\x33\n\x16\x63ross_account_role_arn\x18\x01 \x01(\tR\x13\x63rossAccountRoleArn\x12\x39\n\x19\x63ross_account_external_id\x18\x02 \x01(\tR\x16\x63rossAccountExternalId\x12L\n#cross_account_intermediate_role_arn\x18\x05 \x01(\tR\x1f\x63rossAccountIntermediateRoleArn\x12\x18\n\x07\x65nabled\x18\x03 \x01(\x08R\x07\x65nabled\x12\x32\n\x15\x64\x62\x66s_credentials_path\x18\x04 \x01(\tR\x13\x64\x62\x66sCredentialsPath\"\xeb\x01\n\x10RedisOnlineStore\x12)\n\x10primary_endpoint\x18\x02 \x01(\tR\x0fprimaryEndpoint\x12\x31\n\x14\x61uthentication_token\x18\x04 \x01(\tR\x13\x61uthenticationToken\x12\x1f\n\x0btls_enabled\x18\x06 \x01(\x08R\ntlsEnabled\x12\x18\n\x07\x65nabled\x18\x05 \x01(\x08R\x07\x65nabled\x12&\n\x0finject_host_sni\x18\t \x01(\x08R\rinjectHostSniJ\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\t\"o\n\x13\x42igtableOnlineStore\x12\x18\n\x07\x65nabled\x18\x01 \x01(\x08R\x07\x65nabled\x12\x1d\n\nproject_id\x18\x02 \x01(\tR\tprojectId\x12\x1f\n\x0binstance_id\x18\x03 \x01(\tR\ninstanceId\"\x92\x02\n\x10MonitoringParams\x12%\n\x0euser_specified\x18\x01 \x01(\x08R\ruserSpecified\x12+\n\x11monitor_freshness\x18\x02 \x01(\x08R\x10monitorFreshness\x12W\n\x1a\x65xpected_feature_freshness\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x18\x65xpectedFeatureFreshness\x12\x1f\n\x0b\x61lert_email\x18\x04 \x01(\tR\nalertEmail\x12\x30\n\x14grace_period_seconds\x18\x05 \x01(\x05R\x12gracePeriodSeconds\"\x85\x01\n\x16\x46\x65\x61tureViewEnrichments\x12S\n\x12\x66p_materialization\x18\x04 \x01(\x0b\x32$.tecton_proto.data.FvMaterializationR\x11\x66pMaterializationJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04J\x04\x08\x05\x10\x06\"?\n\rSnowflakeData\x12.\n\x13snowflake_view_name\x18\x01 \x01(\tR\x11snowflakeViewName\"\x81\x01\n\x11\x44\x61taQualityConfig\x12\x30\n\x14\x64\x61ta_quality_enabled\x18\x01 \x01(\x08R\x12\x64\x61taQualityEnabled\x12:\n\x19skip_default_expectations\x18\x02 \x01(\x08R\x17skipDefaultExpectations\"p\n\x18SecondaryKeyOutputColumn\x12@\n\x0btime_window\x18\x01 \x01(\x0b\x32\x1f.tecton_proto.common.TimeWindowR\ntimeWindow\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\"\x92\x02\n\x16\x46\x65\x61tureViewCacheConfig\x12\x1c\n\tnamespace\x18\x01 \x01(\tR\tnamespace\x12(\n\x10\x63\x61\x63he_group_name\x18\x02 \x01(\tR\x0e\x63\x61\x63heGroupName\x12\x41\n\x0fmax_age_seconds\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\rmaxAgeSeconds\x12?\n\x0emax_age_jitter\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0cmaxAgeJitter\x12,\n\x12remapped_join_keys\x18\x05 \x03(\tR\x10remappedJoinKeys*\x95\x01\n\x1aParquetOfflineStoreVersion\x12-\n)PARQUET_OFFLINE_STORE_VERSION_UNSPECIFIED\x10\x00\x12#\n\x1fPARQUET_OFFLINE_STORE_VERSION_1\x10\x01\x12#\n\x1fPARQUET_OFFLINE_STORE_VERSION_2\x10\x02*\x8d\x01\n\x18\x44\x65ltaOfflineStoreVersion\x12+\n\'DELTA_OFFLINE_STORE_VERSION_UNSPECIFIED\x10\x00\x12!\n\x1d\x44\x45LTA_OFFLINE_STORE_VERSION_1\x10\x01\x12!\n\x1d\x44\x45LTA_OFFLINE_STORE_VERSION_2\x10\x02*\xc6\x01\n\x1eMaterializationTimeRangePolicy\x12\x31\n-MATERIALIZATION_TIME_RANGE_POLICY_UNSPECIFIED\x10\x00\x12:\n6MATERIALIZATION_TIME_RANGE_POLICY_FAIL_IF_OUT_OF_RANGE\x10\x01\x12\x35\n1MATERIALIZATION_TIME_RANGE_POLICY_FILTER_TO_RANGE\x10\x02\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.feature_view__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _FEATUREVIEW_OPTIONSENTRY._options = None
  _FEATUREVIEW_OPTIONSENTRY._serialized_options = b'8\001'
  _FEATUREVIEW.fields_by_name['enrichments']._options = None
  _FEATUREVIEW.fields_by_name['enrichments']._serialized_options = b'\030\001'
  _FEATUREVIEW.fields_by_name['framework_version']._options = None
  _FEATUREVIEW.fields_by_name['framework_version']._serialized_options = b'\030\001'
  _FEATUREVIEW.fields_by_name['fw_version']._options = None
  _FEATUREVIEW.fields_by_name['fw_version']._serialized_options = b'\030\001'
  _PARQUETOFFLINESTOREVERSION._serialized_start=10511
  _PARQUETOFFLINESTOREVERSION._serialized_end=10660
  _DELTAOFFLINESTOREVERSION._serialized_start=10663
  _DELTAOFFLINESTOREVERSION._serialized_end=10804
  _MATERIALIZATIONTIMERANGEPOLICY._serialized_start=10807
  _MATERIALIZATIONTIMERANGEPOLICY._serialized_end=11005
  _FEATUREVIEW._serialized_start=826
  _FEATUREVIEW._serialized_end=2972
  _FEATUREVIEW_OPTIONSENTRY._serialized_start=2893
  _FEATUREVIEW_OPTIONSENTRY._serialized_end=2951
  _TEMPORALAGGREGATE._serialized_start=2975
  _TEMPORALAGGREGATE._serialized_end=3463
  _AGGREGATEFEATURE._serialized_start=3466
  _AGGREGATEFEATURE._serialized_end=3938
  _EMBEDDINGFEATURE._serialized_start=3941
  _EMBEDDINGFEATURE._serialized_end=4071
  _INFERENCEFEATURE._serialized_start=4074
  _INFERENCEFEATURE._serialized_end=4317
  _TRAILINGTIMEWINDOWAGGREGATION._serialized_start=4320
  _TRAILINGTIMEWINDOWAGGREGATION._serialized_end=4565
  _TEMPORAL._serialized_start=4568
  _TEMPORAL._serialized_end=5021
  _FEATURETABLE._serialized_start=5024
  _FEATURETABLE._serialized_end=5178
  _REALTIMEFEATUREVIEW._serialized_start=5181
  _REALTIMEFEATUREVIEW._serialized_end=5332
  _FEATUREVIEWSCHEMAS._serialized_start=5335
  _FEATUREVIEWSCHEMAS._serialized_end=5660
  _ONLINESERVINGINDEX._serialized_start=5662
  _ONLINESERVINGINDEX._serialized_end=5711
  _MATERIALIZATIONSTATETRANSITION._serialized_start=5714
  _MATERIALIZATIONSTATETRANSITION._serialized_end=6149
  _PARQUETOFFLINESTOREPARAMS._serialized_start=6152
  _PARQUETOFFLINESTOREPARAMS._serialized_end=6327
  _DELTAOFFLINESTOREPARAMS._serialized_start=6330
  _DELTAOFFLINESTOREPARAMS._serialized_end=6501
  _OFFLINESTOREPARAMS._serialized_start=6504
  _OFFLINESTOREPARAMS._serialized_end=6676
  _FEATUREPUBLISHOFFLINESTORECONFIG._serialized_start=6679
  _FEATUREPUBLISHOFFLINESTORECONFIG._serialized_end=6839
  _NEWMATERIALIZATIONPARAMS._serialized_start=6842
  _NEWMATERIALIZATIONPARAMS._serialized_end=8625
  _ONLINESTOREPARAMS._serialized_start=8628
  _ONLINESTOREPARAMS._serialized_end=8864
  _DYNAMODBONLINESTORE._serialized_start=8867
  _DYNAMODBONLINESTORE._serialized_end=9156
  _REDISONLINESTORE._serialized_start=9159
  _REDISONLINESTORE._serialized_end=9394
  _BIGTABLEONLINESTORE._serialized_start=9396
  _BIGTABLEONLINESTORE._serialized_end=9507
  _MONITORINGPARAMS._serialized_start=9510
  _MONITORINGPARAMS._serialized_end=9784
  _FEATUREVIEWENRICHMENTS._serialized_start=9787
  _FEATUREVIEWENRICHMENTS._serialized_end=9920
  _SNOWFLAKEDATA._serialized_start=9922
  _SNOWFLAKEDATA._serialized_end=9985
  _DATAQUALITYCONFIG._serialized_start=9988
  _DATAQUALITYCONFIG._serialized_end=10117
  _SECONDARYKEYOUTPUTCOLUMN._serialized_start=10119
  _SECONDARYKEYOUTPUTCOLUMN._serialized_end=10231
  _FEATUREVIEWCACHECONFIG._serialized_start=10234
  _FEATUREVIEWCACHECONFIG._serialized_end=10508
# @@protoc_insertion_point(module_scope)



import datetime

from semantic_version import Version

from tecton import _gen_version


def get_version() -> str:
    return _gen_version.VERSION


# Note that get_semantic_version() returns versions that don't exactly match PyPI, e.g. '0.8.0b34' instead of
# '0.8.0-b34'. If you want the same version string as PyPI, use get_version().
#
# NOTE(jake): It's not clear why get_semantic_version() is needed in addition to get_version().
def get_semantic_version() -> str:
    version = get_version()

    _semver = str(Version.coerce(version))
    return _semver


def get_status():
    try:
        from tecton._stamp import BUILD_STATUS

        return BUILD_STATUS
    except ImportError:
        return {}


def get_hash() -> str:
    status = get_status()
    return status.get("GIT_COMMIT", "n/a")


def summary():
    status = get_status()

    ts_seconds = status.get("BUILD_TIMESTAMP", None)
    ts = datetime.datetime.utcfromtimestamp(int(ts_seconds)).isoformat() if ts_seconds else "n/a"

    commit = status.get("GIT_COMMIT", "n/a")

    print(f"Version: {get_version()}")
    print(f"Git Commit: {commit}")
    print(f"Build Datetime: {ts}")
